/**********************************
Triple State Quicksort (Array Based)
Design and code: Ammar Muqaddas
Version 1.0.4 - 30/Apr/2015
ammar.mu@ku.edu.kw
ammar.mu@gmail.com
http://www.solostuff.net/tsqsort/

This is the array based version of the algorithm. It does NOT include further C/C++ related optimization for better element 
swaping. These optimizations are included in the other pointer based (Pointer to Character "PC") version. Therefore, this 
array version will usually run slower. It is provided for simplicity and easiness, for programmers who want to understand the 
algorithm or port it to other languages.

Algorithmically, it is exactly the same as the pointer based version, doing the same number of comparisons and swaps/copies.

Since it's array based, this version depends on the compiler to do the swaping/copying. This means that element size MUST
be known at compile time.
***********************************/

#include <stdlib.h>
#include <memory.h>
#include "Sort.h"
#ifdef RANDOM_SAMPLES
	#include <time.h>
#endif

// Stack implementation
#define PUSH(stackv,value) stackv.pbottom[++stackv.top]=value
#define POP(stackv) stackv.pbottom[stackv.top--]
#define NOTEMPTY(stackv) stackv.top!=-1

// Swap macro
#define SWAP_AR(ar,a1,b1,temp)		\
{	temp = ar[a1];			\
	ar[a1] = ar[b1];		\
	ar[b1] = temp;			\
}

__declspec(thread) sstack Sort::stack1;

void Sort::InitStack(sstack &stackv)
{	stackv.top = -1;
	stackv.pbottom = (int *) malloc(STACK_SIZE*sizeof(int));
}

void Sort::End()
{	free(stack1.pbottom);
	FreeTripleStateQuickSortStorage();
}

inline void Sort::Sort_2(Elem ar[] _DEFINE_CBFUNC)
{	Elem temp;
	
	if (MORE(ar[0],ar[1]))
		SWAP_AR(ar,0,1,temp);
}

inline void Sort::Sort_3(Elem ar[] _DEFINE_CBFUNC)
{	Elem temp;

	if (MORE(ar[0],ar[1]))
	{	
		if (MOREEQ(ar[2],ar[0]))				// 1 0 2
		{	SWAP_AR(ar,0,1,temp);
		}
		else 
		{	
			if (MORE(ar[2],ar[1]))			// 1 2 0
			{	temp = ar[0]; ar[0]=ar[1]; ar[1]=ar[2]; ar[2]=temp; 
			} else							// 2 1 0
			{	SWAP_AR(ar,0,2,temp);
			}
		}
	}
	else
	{	
		if (MOREEQ(ar[2],ar[1])) ;			// 0 1 2
		else
		{	
			if (MOREEQ(ar[2],ar[0]))			// 0 2 1
			{	SWAP_AR(ar,1,2,temp);
			} else
			{	temp = ar[0]; ar[0]=ar[2]; ar[2]=ar[1]; ar[1]=temp; 	// 2 0 1
			}
		}
	}
}

inline void Sort::Sort_4(Elem ar[] _DEFINE_CBFUNC)
{	Elem temp;

	if (MORE(ar[0],ar[1]))
	{	
		if (MOREEQ(ar[2],ar[0]))										// 1 0 2
		{		
			if (MOREEQ(ar[3],ar[0]))
			{	
				if (MOREEQ(ar[3],ar[2]))								// 1 0 2 3
				{	SWAP_AR(ar,0,1,temp);
				}
				else
				{	SWAP_AR(ar,0,1,temp); SWAP_AR(ar,2,3,temp);		// 1 0 3 2
				}
			} else
			{	
				if (MORE(ar[3],ar[1]))
				{	temp=ar[0]; ar[0]=ar[1]; ar[1]=ar[3]; ar[3]=ar[2]; ar[2]=temp; 		// 1 3 0 2
				} else
				{	temp=ar[0]; ar[0]=ar[3]; ar[3]=ar[2]; ar[2]=temp; 					// 3 1 0 2
				}
			}
		}
		else								
		{	
			if (MOREEQ(ar[2],ar[1]))			// 1 2 0
			{	
				if (MORE(ar[3],ar[2]))
				{	
					if (MOREEQ(ar[3],ar[0]))		// 1 2 0 3
					{	temp=ar[0]; ar[0]=ar[1]; ar[1]=ar[2]; ar[2]=temp; 
					} else
					{							// 1 2 3 0
						temp=ar[0]; ar[0]=ar[1]; ar[1]=ar[2]; ar[2]=ar[3]; ar[3]=temp; 	
					}
				} else
				{	
					if (MORE(ar[3],ar[1]))		// 1 3 2 0
					{	temp=ar[0]; ar[0]=ar[1]; ar[1]=ar[3]; ar[3]=temp; 	
					} else
					{	SWAP_AR(ar,0,3,temp);	// 3 1 2 0
					}
				}
			} else							// 2 1 0
			{	
				if (MOREEQ(ar[3],ar[1]))
				{	
					if (MOREEQ(ar[3],ar[0]))		// 2 1 0 3
					{	SWAP_AR(ar,0,2,temp);
					} else
					{							// 2 1 3 0
						temp=ar[0]; ar[0]=ar[2]; ar[2]=ar[3]; ar[3]=temp; 
					}
				} else
				{	
					if (MOREEQ(ar[3],ar[2]))		// 2 3 1 0
					{	temp=ar[0]; ar[0]=ar[2]; ar[2]=ar[1]; ar[1]=ar[3]; ar[3]=temp; 		
					} else
					{	SWAP_AR(ar,0,3,temp); SWAP_AR(ar,1,2,temp);	// 3 2 1 0
					}
				}
			}
		}
	}
	else
	{	////////////////////////////////////////////////////////////////////
		
		if (MOREEQ(ar[2],ar[1]))										// 0 1 2
		{		
			if (MOREEQ(ar[3],ar[1]))									
			{	
				if (MOREEQ(ar[3],ar[2]))								// 0 1 2 3
					;	// Do nothing!!
				else
				{	SWAP_AR(ar,2,3,temp);							// 0 1 3 2
				}
			} else
			{	
				if (MOREEQ(ar[3],ar[0]))								// 0 3 1 2
				{	temp=ar[1]; ar[1]=ar[3]; ar[3]=ar[2]; ar[2]=temp; 
				} else
				{	temp=ar[0]; ar[0]=ar[3]; ar[3]=ar[2]; ar[2]=ar[1]; ar[1]=temp; 		// 3 0 1 2
				}
			}
		}
		else								
		{	
			if (MOREEQ(ar[2],ar[0]))				// 0 2 1
			{	
				if (MOREEQ(ar[3],ar[2]))
				{	
					if (MOREEQ(ar[3],ar[1]))		// 0 2 1 3
					{	SWAP_AR(ar,1,2,temp);
					} else
					{							// 0 2 3 1
						temp=ar[1]; ar[1]=ar[2]; ar[2]=ar[3]; ar[3]=temp; 	
					}
				} else
				{	
					if (MOREEQ(ar[3],ar[0]))		// 0 3 2 1
					{	SWAP_AR(ar,1,3,temp);
					} else
					{	temp=ar[0]; ar[0]=ar[3]; ar[3]=ar[1]; ar[1]=temp; 		// 3 0 2 1
					}
				}
			} else							// 2 0 1
			{	
				if (MOREEQ(ar[3],ar[0]))
				{	
					if (MOREEQ(ar[3],ar[1]))		// 2 0 1 3
					{	temp=ar[0]; ar[0]=ar[2]; ar[2]=ar[1]; ar[1]=temp; 
					} else
					{							// 2 0 3 1
						temp=ar[0]; ar[0]=ar[2]; ar[2]=ar[3]; ar[3]=ar[1]; ar[1]=temp; 	
					}
				} else
				{	
					if (MORE(ar[3],ar[2]))		// 2 3 0 1
					{	SWAP_AR(ar,0,2,temp); SWAP_AR(ar,1,3,temp);
					} else
					{	temp=ar[0]; ar[0]=ar[3]; ar[3]=ar[1]; ar[1]=ar[2]; ar[2]=temp; 		// 3 2 0 1
					}
				}
			}
		}
	}

}

inline short Sort::MedianOf3S(Elem ar[], int a0, int a1, int a2, Elem &piv _DEFINE_CBFUNC)
{	Elem temp;

	if (MOREEQ(ar[a1],ar[a0]))
	{	if (MOREEQ(ar[a2],ar[a0]))
		{	if (MOREEQ(ar[a2],ar[a1]))
			{	// 0 1 2 
				
				piv=ar[a1]; return 1;
			} else
			{	// 0 2 1 
				temp=ar[a1]; ar[a1]=ar[a2]; ar[a2]=temp;
				
			}
		} else
		{	// 2 0 1 
			temp=ar[a0]; ar[a0]=ar[a2]; ar[a2]=ar[a1]; ar[a1]=temp;
			
		}
	} else
	{	if (MOREEQ(ar[a2],ar[a1]))
		{	if (MOREEQ(ar[a2],ar[a0]))
			{	// 1 0 2 
				temp=ar[a0]; ar[a0]=ar[a1]; ar[a1]=temp;
				
			} else
			{	// 1 2 0 
				temp=ar[a0]; ar[a0]=ar[a1]; ar[a1]=ar[a2]; ar[a2]=temp;
				
			}
		} else
		{	// 2 1 0 
			temp=ar[a0]; ar[a0]=ar[a2]; ar[a2]=temp;
			
			piv=ar[a1]; return -1;
		}
	}
	piv=ar[a1]; 
	return 0;
}


inline short Sort::MedianOf5S(Elem ar[], int a0, int a1, int a2, int a3, int a4, Elem &piv _DEFINE_CBFUNC)
{	// Finds the median of 5 by sorting the 5 elements and uses MORE() LESS() macros
	// Returns a hint 1,0,-1 to indicate wether or not the array might be already sorted or reversed.
	
	Elem temp;

	if (MOREEQ(ar[a1],ar[a0]))
	{	if (MOREEQ(ar[a2],ar[a0]))
		{	if (MOREEQ(ar[a2],ar[a1]))
			{	if (MOREEQ(ar[a3],ar[a1]))
				{	if (MOREEQ(ar[a3],ar[a2]))
					{	if (MOREEQ(ar[a4],ar[a1]))
						{	if (MOREEQ(ar[a4],ar[a3]))
							{	// 0 1 2 3 4 
								
								piv=ar[a2]; return 1;
							} else
							{	if (MOREEQ(ar[a4],ar[a2]))
								{	// 0 1 2 4 3 
									temp=ar[a3]; ar[a3]=ar[a4]; ar[a4]=temp;
									
									piv=ar[a2]; return 1;
								} else
								{	// 0 1 4 2 3 
									temp=ar[a2]; ar[a2]=ar[a4]; ar[a4]=ar[a3]; ar[a3]=temp;
									
								}
							}
						} else
						{	if (MOREEQ(ar[a4],ar[a0]))
							{	// 0 4 1 2 3 
								temp=ar[a1]; ar[a1]=ar[a4]; ar[a4]=ar[a3]; ar[a3]=ar[a2]; ar[a2]=temp;
								
							} else
							{	// 4 0 1 2 3 
								temp=ar[a0]; ar[a0]=ar[a4]; ar[a4]=ar[a3]; ar[a3]=ar[a2]; ar[a2]=ar[a1]; ar[a1]=temp;
								
							}
						}
					} else
					{	if (MOREEQ(ar[a4],ar[a1]))
						{	if (MOREEQ(ar[a4],ar[a2]))
							{	// 0 1 3 2 4 
								temp=ar[a2]; ar[a2]=ar[a3]; ar[a3]=temp;
								
							} else
							{	if (MORE(ar[a4],ar[a3]))
								{	// 0 1 3 4 2 
									temp=ar[a2]; ar[a2]=ar[a3]; ar[a3]=ar[a4]; ar[a4]=temp;
									
								} else
								{	// 0 1 4 3 2 
									temp=ar[a2]; ar[a2]=ar[a4]; ar[a4]=temp;
									
								}
							}
						} else
						{	if (MOREEQ(ar[a4],ar[a0]))
							{	// 0 4 1 3 2 
								temp=ar[a1]; ar[a1]=ar[a4]; ar[a4]=ar[a2]; ar[a2]=temp;
								
							} else
							{	// 4 0 1 3 2 
								temp=ar[a0]; ar[a0]=ar[a4]; ar[a4]=ar[a2]; ar[a2]=ar[a1]; ar[a1]=temp;
								
							}
						}
					}
				} else
				{	if (MOREEQ(ar[a3],ar[a0]))
					{	if (MOREEQ(ar[a4],ar[a3]))
						{	if (MOREEQ(ar[a4],ar[a2]))
							{	// 0 3 1 2 4 
								temp=ar[a1]; ar[a1]=ar[a3]; ar[a3]=ar[a2]; ar[a2]=temp;
								
							} else
							{	if (MOREEQ(ar[a4],ar[a1]))
								{	// 0 3 1 4 2 
									temp=ar[a1]; ar[a1]=ar[a3]; ar[a3]=ar[a4]; ar[a4]=ar[a2]; ar[a2]=temp;
									
								} else
								{	// 0 3 4 1 2 
									temp=ar[a1]; ar[a1]=ar[a3]; ar[a3]=temp;
									temp=ar[a2]; ar[a2]=ar[a4]; ar[a4]=temp;
									
								}
							}
						} else
						{	if (MOREEQ(ar[a4],ar[a0]))
							{	// 0 4 3 1 2 
								temp=ar[a1]; ar[a1]=ar[a4]; ar[a4]=ar[a2]; ar[a2]=ar[a3]; ar[a3]=temp;
								
							} else
							{	// 4 0 3 1 2 
								temp=ar[a0]; ar[a0]=ar[a4]; ar[a4]=ar[a2]; ar[a2]=ar[a3]; ar[a3]=ar[a1]; ar[a1]=temp;
								
							}
						}
					} else
					{	if (MOREEQ(ar[a4],ar[a0]))
						{	if (MOREEQ(ar[a4],ar[a2]))
							{	// 3 0 1 2 4 
								temp=ar[a0]; ar[a0]=ar[a3]; ar[a3]=ar[a2]; ar[a2]=ar[a1]; ar[a1]=temp;
								
							} else
							{	if (MOREEQ(ar[a4],ar[a1]))
								{	// 3 0 1 4 2 
									temp=ar[a0]; ar[a0]=ar[a3]; ar[a3]=ar[a4]; ar[a4]=ar[a2]; ar[a2]=ar[a1]; ar[a1]=temp;
									
								} else
								{	// 3 0 4 1 2 
									temp=ar[a0]; ar[a0]=ar[a3]; ar[a3]=ar[a1]; ar[a1]=temp;
									temp=ar[a2]; ar[a2]=ar[a4]; ar[a4]=temp;
									
								}
							}
						} else
						{	if (MOREEQ(ar[a4],ar[a3]))
							{	// 3 4 0 1 2 
								temp=ar[a0]; ar[a0]=ar[a3]; ar[a3]=ar[a1]; ar[a1]=ar[a4]; ar[a4]=ar[a2]; ar[a2]=temp;
								
							} else
							{	// 4 3 0 1 2 
								temp=ar[a0]; ar[a0]=ar[a4]; ar[a4]=ar[a2]; ar[a2]=temp;
								temp=ar[a1]; ar[a1]=ar[a3]; ar[a3]=temp;
								
							}
						}
					}
				}
			} else
			{	if (MOREEQ(ar[a3],ar[a2]))
				{	if (MOREEQ(ar[a3],ar[a1]))
					{	if (MOREEQ(ar[a4],ar[a2]))
						{	if (MOREEQ(ar[a4],ar[a3]))
							{	// 0 2 1 3 4 
								temp=ar[a1]; ar[a1]=ar[a2]; ar[a2]=temp;
								
							} else
							{	if (MORE(ar[a4],ar[a1]))
								{	// 0 2 1 4 3 
									temp=ar[a1]; ar[a1]=ar[a2]; ar[a2]=temp;
									temp=ar[a3]; ar[a3]=ar[a4]; ar[a4]=temp;
									
								} else
								{	// 0 2 4 1 3 
									temp=ar[a1]; ar[a1]=ar[a2]; ar[a2]=ar[a4]; ar[a4]=ar[a3]; ar[a3]=temp;
									
								}
							}
						} else
						{	if (MOREEQ(ar[a4],ar[a0]))
							{	// 0 4 2 1 3 
								temp=ar[a1]; ar[a1]=ar[a4]; ar[a4]=ar[a3]; ar[a3]=temp;
								
							} else
							{	// 4 0 2 1 3 
								temp=ar[a0]; ar[a0]=ar[a4]; ar[a4]=ar[a3]; ar[a3]=ar[a1]; ar[a1]=temp;
								
							}
						}
					} else
					{	if (MORE(ar[a4],ar[a2]))
						{	if (MOREEQ(ar[a4],ar[a1]))
							{	// 0 2 3 1 4 
								temp=ar[a1]; ar[a1]=ar[a2]; ar[a2]=ar[a3]; ar[a3]=temp;
								
							} else
							{	if (MORE(ar[a4],ar[a3]))
								{	// 0 2 3 4 1 
									temp=ar[a1]; ar[a1]=ar[a2]; ar[a2]=ar[a3]; ar[a3]=ar[a4]; ar[a4]=temp;
									
								} else
								{	// 0 2 4 3 1 
									temp=ar[a1]; ar[a1]=ar[a2]; ar[a2]=ar[a4]; ar[a4]=temp;
									
								}
							}
						} else
						{	if (MOREEQ(ar[a4],ar[a0]))
							{	// 0 4 2 3 1 
								temp=ar[a1]; ar[a1]=ar[a4]; ar[a4]=temp;
								
							} else
							{	// 4 0 2 3 1 
								temp=ar[a0]; ar[a0]=ar[a4]; ar[a4]=ar[a1]; ar[a1]=temp;
								
							}
						}
					}
				} else
				{	if (MOREEQ(ar[a3],ar[a0]))
					{	if (MOREEQ(ar[a4],ar[a3]))
						{	if (MOREEQ(ar[a4],ar[a1]))
							{	// 0 3 2 1 4 
								temp=ar[a1]; ar[a1]=ar[a3]; ar[a3]=temp;
								
							} else
							{	if (MOREEQ(ar[a4],ar[a2]))
								{	// 0 3 2 4 1 
									temp=ar[a1]; ar[a1]=ar[a3]; ar[a3]=ar[a4]; ar[a4]=temp;
									
								} else
								{	// 0 3 4 2 1 
									temp=ar[a1]; ar[a1]=ar[a3]; ar[a3]=ar[a2]; ar[a2]=ar[a4]; ar[a4]=temp;
									
								}
							}
						} else
						{	if (MOREEQ(ar[a4],ar[a0]))
							{	// 0 4 3 2 1 
								temp=ar[a1]; ar[a1]=ar[a4]; ar[a4]=temp;
								temp=ar[a2]; ar[a2]=ar[a3]; ar[a3]=temp;
								
							} else
							{	// 4 0 3 2 1 
								temp=ar[a0]; ar[a0]=ar[a4]; ar[a4]=ar[a1]; ar[a1]=temp;
								temp=ar[a2]; ar[a2]=ar[a3]; ar[a3]=temp;
								
							}
						}
					} else
					{	if (MOREEQ(ar[a4],ar[a0]))
						{	if (MOREEQ(ar[a4],ar[a1]))
							{	// 3 0 2 1 4 
								temp=ar[a0]; ar[a0]=ar[a3]; ar[a3]=ar[a1]; ar[a1]=temp;
								
							} else
							{	if (MOREEQ(ar[a4],ar[a2]))
								{	// 3 0 2 4 1 
									temp=ar[a0]; ar[a0]=ar[a3]; ar[a3]=ar[a4]; ar[a4]=ar[a1]; ar[a1]=temp;
									
								} else
								{	// 3 0 4 2 1 
									temp=ar[a0]; ar[a0]=ar[a3]; ar[a3]=ar[a2]; ar[a2]=ar[a4]; ar[a4]=ar[a1]; ar[a1]=temp;
									
								}
							}
						} else
						{	if (MORE(ar[a4],ar[a3]))
							{	// 3 4 0 2 1 
								temp=ar[a0]; ar[a0]=ar[a3]; ar[a3]=ar[a2]; ar[a2]=temp;
								temp=ar[a1]; ar[a1]=ar[a4]; ar[a4]=temp;
								
							} else
							{	// 4 3 0 2 1 
								temp=ar[a0]; ar[a0]=ar[a4]; ar[a4]=ar[a1]; ar[a1]=ar[a3]; ar[a3]=ar[a2]; ar[a2]=temp;
								
							}
						}
					}
				}
			}
		} else
		{	if (MOREEQ(ar[a3],ar[a0]))
			{	if (MOREEQ(ar[a3],ar[a1]))
				{	if (MOREEQ(ar[a4],ar[a0]))
					{	if (MOREEQ(ar[a4],ar[a3]))
						{	// 2 0 1 3 4 
							temp=ar[a0]; ar[a0]=ar[a2]; ar[a2]=ar[a1]; ar[a1]=temp;
							
						} else
						{	if (MORE(ar[a4],ar[a1]))
							{	// 2 0 1 4 3 
								temp=ar[a0]; ar[a0]=ar[a2]; ar[a2]=ar[a1]; ar[a1]=temp;
								temp=ar[a3]; ar[a3]=ar[a4]; ar[a4]=temp;
								
							} else
							{	// 2 0 4 1 3 
								temp=ar[a0]; ar[a0]=ar[a2]; ar[a2]=ar[a4]; ar[a4]=ar[a3]; ar[a3]=ar[a1]; ar[a1]=temp;
								
							}
						}
					} else
					{	if (MORE(ar[a4],ar[a2]))
						{	// 2 4 0 1 3 
							temp=ar[a0]; ar[a0]=ar[a2]; ar[a2]=temp;
							temp=ar[a1]; ar[a1]=ar[a4]; ar[a4]=ar[a3]; ar[a3]=temp;
							
						} else
						{	// 4 2 0 1 3 
							temp=ar[a0]; ar[a0]=ar[a4]; ar[a4]=ar[a3]; ar[a3]=ar[a1]; ar[a1]=ar[a2]; ar[a2]=temp;
							
						}
					}
				} else
				{	if (MOREEQ(ar[a4],ar[a0]))
					{	if (MOREEQ(ar[a4],ar[a1]))
						{	// 2 0 3 1 4 
							temp=ar[a0]; ar[a0]=ar[a2]; ar[a2]=ar[a3]; ar[a3]=ar[a1]; ar[a1]=temp;
							
						} else
						{	if (MORE(ar[a4],ar[a3]))
							{	// 2 0 3 4 1 
								temp=ar[a0]; ar[a0]=ar[a2]; ar[a2]=ar[a3]; ar[a3]=ar[a4]; ar[a4]=ar[a1]; ar[a1]=temp;
								
							} else
							{	// 2 0 4 3 1 
								temp=ar[a0]; ar[a0]=ar[a2]; ar[a2]=ar[a4]; ar[a4]=ar[a1]; ar[a1]=temp;
								
							}
						}
					} else
					{	if (MORE(ar[a4],ar[a2]))
						{	// 2 4 0 3 1 
							temp=ar[a0]; ar[a0]=ar[a2]; ar[a2]=temp;
							temp=ar[a1]; ar[a1]=ar[a4]; ar[a4]=temp;
							
						} else
						{	// 4 2 0 3 1 
							temp=ar[a0]; ar[a0]=ar[a4]; ar[a4]=ar[a1]; ar[a1]=ar[a2]; ar[a2]=temp;
							
						}
					}
				}
			} else
			{	if (MORE(ar[a3],ar[a2]))
				{	if (MOREEQ(ar[a4],ar[a3]))
					{	if (MOREEQ(ar[a4],ar[a1]))
						{	// 2 3 0 1 4 
							temp=ar[a0]; ar[a0]=ar[a2]; ar[a2]=temp;
							temp=ar[a1]; ar[a1]=ar[a3]; ar[a3]=temp;
							
						} else
						{	if (MORE(ar[a4],ar[a0]))
							{	// 2 3 0 4 1 
								temp=ar[a0]; ar[a0]=ar[a2]; ar[a2]=temp;
								temp=ar[a1]; ar[a1]=ar[a3]; ar[a3]=ar[a4]; ar[a4]=temp;
								
							} else
							{	// 2 3 4 0 1 
								temp=ar[a0]; ar[a0]=ar[a2]; ar[a2]=ar[a4]; ar[a4]=ar[a1]; ar[a1]=ar[a3]; ar[a3]=temp;
								
							}
						}
					} else
					{	if (MORE(ar[a4],ar[a2]))
						{	// 2 4 3 0 1 
							temp=ar[a0]; ar[a0]=ar[a2]; ar[a2]=ar[a3]; ar[a3]=temp;
							temp=ar[a1]; ar[a1]=ar[a4]; ar[a4]=temp;
							
						} else
						{	// 4 2 3 0 1 
							temp=ar[a0]; ar[a0]=ar[a4]; ar[a4]=ar[a1]; ar[a1]=ar[a2]; ar[a2]=ar[a3]; ar[a3]=temp;
							
						}
					}
				} else
				{	if (MOREEQ(ar[a4],ar[a2]))
					{	if (MOREEQ(ar[a4],ar[a1]))
						{	// 3 2 0 1 4 
							temp=ar[a0]; ar[a0]=ar[a3]; ar[a3]=ar[a1]; ar[a1]=ar[a2]; ar[a2]=temp;
							
						} else
						{	if (MOREEQ(ar[a4],ar[a0]))
							{	// 3 2 0 4 1 
								temp=ar[a0]; ar[a0]=ar[a3]; ar[a3]=ar[a4]; ar[a4]=ar[a1]; ar[a1]=ar[a2]; ar[a2]=temp;
								
							} else
							{	// 3 2 4 0 1 
								temp=ar[a0]; ar[a0]=ar[a3]; ar[a3]=temp;
								temp=ar[a1]; ar[a1]=ar[a2]; ar[a2]=ar[a4]; ar[a4]=temp;
								
							}
						}
					} else
					{	if (MORE(ar[a4],ar[a3]))
						{	// 3 4 2 0 1 
							temp=ar[a0]; ar[a0]=ar[a3]; ar[a3]=temp;
							temp=ar[a1]; ar[a1]=ar[a4]; ar[a4]=temp;
							
							piv=ar[a2]; return -1;
						} else
						{	// 4 3 2 0 1 
							temp=ar[a0]; ar[a0]=ar[a4]; ar[a4]=ar[a1]; ar[a1]=ar[a3]; ar[a3]=temp;
							
							piv=ar[a2]; return -1;
						}
					}
				}
			}
		}
	} else
	{	if (MOREEQ(ar[a2],ar[a1]))
		{	if (MOREEQ(ar[a2],ar[a0]))
			{	if (MOREEQ(ar[a3],ar[a0]))
				{	if (MOREEQ(ar[a3],ar[a2]))
					{	if (MOREEQ(ar[a4],ar[a0]))
						{	if (MOREEQ(ar[a4],ar[a3]))
							{	// 1 0 2 3 4 
								temp=ar[a0]; ar[a0]=ar[a1]; ar[a1]=temp;
								
								piv=ar[a2]; return 1;
							} else
							{	if (MOREEQ(ar[a4],ar[a2]))
								{	// 1 0 2 4 3 
									temp=ar[a0]; ar[a0]=ar[a1]; ar[a1]=temp;
									temp=ar[a3]; ar[a3]=ar[a4]; ar[a4]=temp;
									
									piv=ar[a2]; return 1;
								} else
								{	// 1 0 4 2 3 
									temp=ar[a0]; ar[a0]=ar[a1]; ar[a1]=temp;
									temp=ar[a2]; ar[a2]=ar[a4]; ar[a4]=ar[a3]; ar[a3]=temp;
									
								}
							}
						} else
						{	if (MORE(ar[a4],ar[a1]))
							{	// 1 4 0 2 3 
								temp=ar[a0]; ar[a0]=ar[a1]; ar[a1]=ar[a4]; ar[a4]=ar[a3]; ar[a3]=ar[a2]; ar[a2]=temp;
								
							} else
							{	// 4 1 0 2 3 
								temp=ar[a0]; ar[a0]=ar[a4]; ar[a4]=ar[a3]; ar[a3]=ar[a2]; ar[a2]=temp;
								
							}
						}
					} else
					{	if (MORE(ar[a4],ar[a0]))
						{	if (MOREEQ(ar[a4],ar[a2]))
							{	// 1 0 3 2 4 
								temp=ar[a0]; ar[a0]=ar[a1]; ar[a1]=temp;
								temp=ar[a2]; ar[a2]=ar[a3]; ar[a3]=temp;
								
							} else
							{	if (MORE(ar[a4],ar[a3]))
								{	// 1 0 3 4 2 
									temp=ar[a0]; ar[a0]=ar[a1]; ar[a1]=temp;
									temp=ar[a2]; ar[a2]=ar[a3]; ar[a3]=ar[a4]; ar[a4]=temp;
									
								} else
								{	// 1 0 4 3 2 
									temp=ar[a0]; ar[a0]=ar[a1]; ar[a1]=temp;
									temp=ar[a2]; ar[a2]=ar[a4]; ar[a4]=temp;
									
								}
							}
						} else
						{	if (MORE(ar[a4],ar[a1]))
							{	// 1 4 0 3 2 
								temp=ar[a0]; ar[a0]=ar[a1]; ar[a1]=ar[a4]; ar[a4]=ar[a2]; ar[a2]=temp;
								
							} else
							{	// 4 1 0 3 2 
								temp=ar[a0]; ar[a0]=ar[a4]; ar[a4]=ar[a2]; ar[a2]=temp;
								
							}
						}
					}
				} else
				{	if (MORE(ar[a3],ar[a1]))
					{	if (MOREEQ(ar[a4],ar[a3]))
						{	if (MOREEQ(ar[a4],ar[a2]))
							{	// 1 3 0 2 4 
								temp=ar[a0]; ar[a0]=ar[a1]; ar[a1]=ar[a3]; ar[a3]=ar[a2]; ar[a2]=temp;
								
							} else
							{	if (MOREEQ(ar[a4],ar[a0]))
								{	// 1 3 0 4 2 
									temp=ar[a0]; ar[a0]=ar[a1]; ar[a1]=ar[a3]; ar[a3]=ar[a4]; ar[a4]=ar[a2]; ar[a2]=temp;
									
								} else
								{	// 1 3 4 0 2 
									temp=ar[a0]; ar[a0]=ar[a1]; ar[a1]=ar[a3]; ar[a3]=temp;
									temp=ar[a2]; ar[a2]=ar[a4]; ar[a4]=temp;
									
								}
							}
						} else
						{	if (MORE(ar[a4],ar[a1]))
							{	// 1 4 3 0 2 
								temp=ar[a0]; ar[a0]=ar[a1]; ar[a1]=ar[a4]; ar[a4]=ar[a2]; ar[a2]=ar[a3]; ar[a3]=temp;
								
							} else
							{	// 4 1 3 0 2 
								temp=ar[a0]; ar[a0]=ar[a4]; ar[a4]=ar[a2]; ar[a2]=ar[a3]; ar[a3]=temp;
								
							}
						}
					} else
					{	if (MOREEQ(ar[a4],ar[a1]))
						{	if (MOREEQ(ar[a4],ar[a2]))
							{	// 3 1 0 2 4 
								temp=ar[a0]; ar[a0]=ar[a3]; ar[a3]=ar[a2]; ar[a2]=temp;
								
							} else
							{	if (MOREEQ(ar[a4],ar[a0]))
								{	// 3 1 0 4 2 
									temp=ar[a0]; ar[a0]=ar[a3]; ar[a3]=ar[a4]; ar[a4]=ar[a2]; ar[a2]=temp;
									
								} else
								{	// 3 1 4 0 2 
									temp=ar[a0]; ar[a0]=ar[a3]; ar[a3]=temp;
									temp=ar[a2]; ar[a2]=ar[a4]; ar[a4]=temp;
									
								}
							}
						} else
						{	if (MORE(ar[a4],ar[a3]))
							{	// 3 4 1 0 2 
								temp=ar[a0]; ar[a0]=ar[a3]; ar[a3]=temp;
								temp=ar[a1]; ar[a1]=ar[a4]; ar[a4]=ar[a2]; ar[a2]=temp;
								
							} else
							{	// 4 3 1 0 2 
								temp=ar[a0]; ar[a0]=ar[a4]; ar[a4]=ar[a2]; ar[a2]=ar[a1]; ar[a1]=ar[a3]; ar[a3]=temp;
								
							}
						}
					}
				}
			} else
			{	if (MOREEQ(ar[a3],ar[a2]))
				{	if (MORE(ar[a3],ar[a0]))
					{	if (MOREEQ(ar[a4],ar[a2]))
						{	if (MOREEQ(ar[a4],ar[a3]))
							{	// 1 2 0 3 4 
								temp=ar[a0]; ar[a0]=ar[a1]; ar[a1]=ar[a2]; ar[a2]=temp;
								
							} else
							{	if (MORE(ar[a4],ar[a0]))
								{	// 1 2 0 4 3 
									temp=ar[a0]; ar[a0]=ar[a1]; ar[a1]=ar[a2]; ar[a2]=temp;
									temp=ar[a3]; ar[a3]=ar[a4]; ar[a4]=temp;
									
								} else
								{	// 1 2 4 0 3 
									temp=ar[a0]; ar[a0]=ar[a1]; ar[a1]=ar[a2]; ar[a2]=ar[a4]; ar[a4]=ar[a3]; ar[a3]=temp;
									
								}
							}
						} else
						{	if (MORE(ar[a4],ar[a1]))
							{	// 1 4 2 0 3 
								temp=ar[a0]; ar[a0]=ar[a1]; ar[a1]=ar[a4]; ar[a4]=ar[a3]; ar[a3]=temp;
								
							} else
							{	// 4 1 2 0 3 
								temp=ar[a0]; ar[a0]=ar[a4]; ar[a4]=ar[a3]; ar[a3]=temp;
								
							}
						}
					} else
					{	if (MORE(ar[a4],ar[a2]))
						{	if (MOREEQ(ar[a4],ar[a0]))
							{	// 1 2 3 0 4 
								temp=ar[a0]; ar[a0]=ar[a1]; ar[a1]=ar[a2]; ar[a2]=ar[a3]; ar[a3]=temp;
								
							} else
							{	if (MORE(ar[a4],ar[a3]))
								{	// 1 2 3 4 0 
									temp=ar[a0]; ar[a0]=ar[a1]; ar[a1]=ar[a2]; ar[a2]=ar[a3]; ar[a3]=ar[a4]; ar[a4]=temp;
									
								} else
								{	// 1 2 4 3 0 
									temp=ar[a0]; ar[a0]=ar[a1]; ar[a1]=ar[a2]; ar[a2]=ar[a4]; ar[a4]=temp;
									
								}
							}
						} else
						{	if (MORE(ar[a4],ar[a1]))
							{	// 1 4 2 3 0 
								temp=ar[a0]; ar[a0]=ar[a1]; ar[a1]=ar[a4]; ar[a4]=temp;
								
							} else
							{	// 4 1 2 3 0 
								temp=ar[a0]; ar[a0]=ar[a4]; ar[a4]=temp;
								
							}
						}
					}
				} else
				{	if (MORE(ar[a3],ar[a1]))
					{	if (MOREEQ(ar[a4],ar[a3]))
						{	if (MOREEQ(ar[a4],ar[a0]))
							{	// 1 3 2 0 4 
								temp=ar[a0]; ar[a0]=ar[a1]; ar[a1]=ar[a3]; ar[a3]=temp;
								
							} else
							{	if (MOREEQ(ar[a4],ar[a2]))
								{	// 1 3 2 4 0 
									temp=ar[a0]; ar[a0]=ar[a1]; ar[a1]=ar[a3]; ar[a3]=ar[a4]; ar[a4]=temp;
									
								} else
								{	// 1 3 4 2 0 
									temp=ar[a0]; ar[a0]=ar[a1]; ar[a1]=ar[a3]; ar[a3]=ar[a2]; ar[a2]=ar[a4]; ar[a4]=temp;
									
								}
							}
						} else
						{	if (MORE(ar[a4],ar[a1]))
							{	// 1 4 3 2 0 
								temp=ar[a0]; ar[a0]=ar[a1]; ar[a1]=ar[a4]; ar[a4]=temp;
								temp=ar[a2]; ar[a2]=ar[a3]; ar[a3]=temp;
								
							} else
							{	// 4 1 3 2 0 
								temp=ar[a0]; ar[a0]=ar[a4]; ar[a4]=temp;
								temp=ar[a2]; ar[a2]=ar[a3]; ar[a3]=temp;
								
							}
						}
					} else
					{	if (MOREEQ(ar[a4],ar[a1]))
						{	if (MOREEQ(ar[a4],ar[a0]))
							{	// 3 1 2 0 4 
								temp=ar[a0]; ar[a0]=ar[a3]; ar[a3]=temp;
								
							} else
							{	if (MOREEQ(ar[a4],ar[a2]))
								{	// 3 1 2 4 0 
									temp=ar[a0]; ar[a0]=ar[a3]; ar[a3]=ar[a4]; ar[a4]=temp;
									
								} else
								{	// 3 1 4 2 0 
									temp=ar[a0]; ar[a0]=ar[a3]; ar[a3]=ar[a2]; ar[a2]=ar[a4]; ar[a4]=temp;
									
								}
							}
						} else
						{	if (MOREEQ(ar[a4],ar[a3]))
							{	// 3 4 1 2 0 
								temp=ar[a0]; ar[a0]=ar[a3]; ar[a3]=ar[a2]; ar[a2]=ar[a1]; ar[a1]=ar[a4]; ar[a4]=temp;
								
							} else
							{	// 4 3 1 2 0 
								temp=ar[a0]; ar[a0]=ar[a4]; ar[a4]=temp;
								temp=ar[a1]; ar[a1]=ar[a3]; ar[a3]=ar[a2]; ar[a2]=temp;
								
							}
						}
					}
				}
			}
		} else
		{	if (MOREEQ(ar[a3],ar[a1]))
			{	if (MOREEQ(ar[a3],ar[a0]))
				{	if (MOREEQ(ar[a4],ar[a1]))
					{	if (MOREEQ(ar[a4],ar[a3]))
						{	// 2 1 0 3 4 
							temp=ar[a0]; ar[a0]=ar[a2]; ar[a2]=temp;
							
						} else
						{	if (MORE(ar[a4],ar[a0]))
							{	// 2 1 0 4 3 
								temp=ar[a0]; ar[a0]=ar[a2]; ar[a2]=temp;
								temp=ar[a3]; ar[a3]=ar[a4]; ar[a4]=temp;
								
							} else
							{	// 2 1 4 0 3 
								temp=ar[a0]; ar[a0]=ar[a2]; ar[a2]=ar[a4]; ar[a4]=ar[a3]; ar[a3]=temp;
								
							}
						}
					} else
					{	if (MOREEQ(ar[a4],ar[a2]))
						{	// 2 4 1 0 3 
							temp=ar[a0]; ar[a0]=ar[a2]; ar[a2]=ar[a1]; ar[a1]=ar[a4]; ar[a4]=ar[a3]; ar[a3]=temp;
							
						} else
						{	// 4 2 1 0 3 
							temp=ar[a0]; ar[a0]=ar[a4]; ar[a4]=ar[a3]; ar[a3]=temp;
							temp=ar[a1]; ar[a1]=ar[a2]; ar[a2]=temp;
							
						}
					}
				} else
				{	if (MOREEQ(ar[a4],ar[a1]))
					{	if (MOREEQ(ar[a4],ar[a0]))
						{	// 2 1 3 0 4 
							temp=ar[a0]; ar[a0]=ar[a2]; ar[a2]=ar[a3]; ar[a3]=temp;
							
						} else
						{	if (MORE(ar[a4],ar[a3]))
							{	// 2 1 3 4 0 
								temp=ar[a0]; ar[a0]=ar[a2]; ar[a2]=ar[a3]; ar[a3]=ar[a4]; ar[a4]=temp;
								
							} else
							{	// 2 1 4 3 0 
								temp=ar[a0]; ar[a0]=ar[a2]; ar[a2]=ar[a4]; ar[a4]=temp;
								
							}
						}
					} else
					{	if (MOREEQ(ar[a4],ar[a2]))
						{	// 2 4 1 3 0 
							temp=ar[a0]; ar[a0]=ar[a2]; ar[a2]=ar[a1]; ar[a1]=ar[a4]; ar[a4]=temp;
							
						} else
						{	// 4 2 1 3 0 
							temp=ar[a0]; ar[a0]=ar[a4]; ar[a4]=temp;
							temp=ar[a1]; ar[a1]=ar[a2]; ar[a2]=temp;
							
						}
					}
				}
			} else
			{	if (MOREEQ(ar[a3],ar[a2]))
				{	if (MOREEQ(ar[a4],ar[a3]))
					{	if (MOREEQ(ar[a4],ar[a0]))
						{	// 2 3 1 0 4 
							temp=ar[a0]; ar[a0]=ar[a2]; ar[a2]=ar[a1]; ar[a1]=ar[a3]; ar[a3]=temp;
							
						} else
						{	if (MOREEQ(ar[a4],ar[a1]))
							{	// 2 3 1 4 0 
								temp=ar[a0]; ar[a0]=ar[a2]; ar[a2]=ar[a1]; ar[a1]=ar[a3]; ar[a3]=ar[a4]; ar[a4]=temp;
								
							} else
							{	// 2 3 4 1 0 
								temp=ar[a0]; ar[a0]=ar[a2]; ar[a2]=ar[a4]; ar[a4]=temp;
								temp=ar[a1]; ar[a1]=ar[a3]; ar[a3]=temp;
								
							}
						}
					} else
					{	if (MOREEQ(ar[a4],ar[a2]))
						{	// 2 4 3 1 0 
							temp=ar[a0]; ar[a0]=ar[a2]; ar[a2]=ar[a3]; ar[a3]=ar[a1]; ar[a1]=ar[a4]; ar[a4]=temp;
							
						} else
						{	// 4 2 3 1 0 
							temp=ar[a0]; ar[a0]=ar[a4]; ar[a4]=temp;
							temp=ar[a1]; ar[a1]=ar[a2]; ar[a2]=ar[a3]; ar[a3]=temp;
							
						}
					}
				} else
				{	if (MORE(ar[a4],ar[a2]))
					{	if (MOREEQ(ar[a4],ar[a0]))
						{	// 3 2 1 0 4 
							temp=ar[a0]; ar[a0]=ar[a3]; ar[a3]=temp;
							temp=ar[a1]; ar[a1]=ar[a2]; ar[a2]=temp;
							
						} else
						{	if (MORE(ar[a4],ar[a1]))
							{	// 3 2 1 4 0 
								temp=ar[a0]; ar[a0]=ar[a3]; ar[a3]=ar[a4]; ar[a4]=temp;
								temp=ar[a1]; ar[a1]=ar[a2]; ar[a2]=temp;
								
							} else
							{	// 3 2 4 1 0 
								temp=ar[a0]; ar[a0]=ar[a3]; ar[a3]=ar[a1]; ar[a1]=ar[a2]; ar[a2]=ar[a4]; ar[a4]=temp;
								
							}
						}
					} else
					{	if (MOREEQ(ar[a4],ar[a3]))
						{	// 3 4 2 1 0 
							temp=ar[a0]; ar[a0]=ar[a3]; ar[a3]=ar[a1]; ar[a1]=ar[a4]; ar[a4]=temp;
							
							piv=ar[a2]; return -1;
						} else
						{	// 4 3 2 1 0 
							temp=ar[a0]; ar[a0]=ar[a4]; ar[a4]=temp;
							temp=ar[a1]; ar[a1]=ar[a3]; ar[a3]=temp;
							
							piv=ar[a2]; return -1;
						}
					}
				}
			}
		}
	}
	piv=ar[a2]; 
	return 0;
}


inline short Sort::MedianOf3NS(Elem ar[], int a0, int a1, int a2, int &ipiv _DEFINE_CBFUNC)
{	
	if (MORE(ar[a0],ar[a1]))
		if (MOREEQ(ar[a2],ar[a0])) // a1 a0 a2
			ipiv=a0;
		else
		if (MORE(ar[a2],ar[a1])) //a1 a2 a0
			ipiv=a2;
		else //a2 a1 a0
		{	ipiv=a1; return -1;
		}
	else 
	if (MOREEQ(ar[a2],ar[a1])) //a0 a1 a2
	{	ipiv=a1; return 1;
	} else
	if (MOREEQ(ar[a2],ar[a0])) //a0 a2 a1
		ipiv=a2;
	else  //a2 a0 a1
		ipiv=a0;
	return 0;
}


inline int Sort::MedOfThreeNoSort(Elem ar[], int a, int b, int c _DEFINE_CBFUNC)
{	// Finds the median of three without sorting the three elements and uses MORE() LESS() macros

	
	if (MORE(ar[a],ar[b]))
	{	
		if (MOREEQ(ar[c],ar[a])) // b a c
			return a;
		else 
		{	
			if (MORE(ar[c],ar[b])) //b c a
				return c;
			else //c b a
				return b;
		}
	}
	else 
	{	
		if (MOREEQ(ar[c],ar[b])) //a b c
			return b;
		else 
		{	
			if (MOREEQ(ar[c],ar[a])) //a c b
				return c;
			else  //c a b
				return a;
		}
	}
}


inline short Sort::MedianOf5NS(Elem ar[], int a0, int a1, int a2, int a3, int a4, int &ipiv _DEFINE_CBFUNC)
{	// Finds the median of 5 WITHOUT sorting the 5 elements and uses MORE() LESS() macros
	// Returns a hint 1,0,-1 to indicate wether or not the array might be already sorted or reversed.

	if (MOREEQ(ar[a1],ar[a0]))
	{	if (MOREEQ(ar[a2],ar[a0]))
		{	if (MOREEQ(ar[a2],ar[a1]))
			{	if (MOREEQ(ar[a3],ar[a1]))
				{	if (MOREEQ(ar[a3],ar[a2]))
					{	if (MOREEQ(ar[a4],ar[a1]))
						{	if (MOREEQ(ar[a4],ar[a3]))
							{	// 0 1 2 3 4 
								ipiv=a2;
								return 1;
							} else
							{	if (MOREEQ(ar[a4],ar[a2]))
								{	// 0 1 2 4 3 
									ipiv=a2;
									return 1;
								} else
								{	// 0 1 4 2 3 
									ipiv=a4;
								}
							}
						} else
						{	if (MOREEQ(ar[a4],ar[a0]))
							{	// 0 4 1 2 3 
								ipiv=a1;
							} else
							{	// 4 0 1 2 3 
								ipiv=a1;
							}
						}
					} else
					{	if (MOREEQ(ar[a4],ar[a1]))
						{	if (MOREEQ(ar[a4],ar[a2]))
							{	// 0 1 3 2 4 
								ipiv=a3;
							} else
							{	if (MORE(ar[a4],ar[a3]))
								{	// 0 1 3 4 2 
									ipiv=a3;
								} else
								{	// 0 1 4 3 2 
									ipiv=a4;
								}
							}
						} else
						{	if (MOREEQ(ar[a4],ar[a0]))
							{	// 0 4 1 3 2 
								ipiv=a1;
							} else
							{	// 4 0 1 3 2 
								ipiv=a1;
							}
						}
					}
				} else
				{	if (MOREEQ(ar[a3],ar[a0]))
					{	if (MOREEQ(ar[a4],ar[a3]))
						{	if (MOREEQ(ar[a4],ar[a2]))
							{	// 0 3 1 2 4 
								ipiv=a1;
							} else
							{	if (MOREEQ(ar[a4],ar[a1]))
								{	// 0 3 1 4 2 
									ipiv=a1;
								} else
								{	// 0 3 4 1 2 
									ipiv=a4;
								}
							}
						} else
						{	if (MOREEQ(ar[a4],ar[a0]))
							{	// 0 4 3 1 2 
								ipiv=a3;
							} else
							{	// 4 0 3 1 2 
								ipiv=a3;
							}
						}
					} else
					{	if (MOREEQ(ar[a4],ar[a0]))
						{	if (MOREEQ(ar[a4],ar[a2]))
							{	// 3 0 1 2 4 
								ipiv=a1;
							} else
							{	if (MOREEQ(ar[a4],ar[a1]))
								{	// 3 0 1 4 2 
									ipiv=a1;
								} else
								{	// 3 0 4 1 2 
									ipiv=a4;
								}
							}
						} else
						{	if (MOREEQ(ar[a4],ar[a3]))
							{	// 3 4 0 1 2 
								ipiv=a0;
							} else
							{	// 4 3 0 1 2 
								ipiv=a0;
							}
						}
					}
				}
			} else
			{	if (MOREEQ(ar[a3],ar[a2]))
				{	if (MOREEQ(ar[a3],ar[a1]))
					{	if (MOREEQ(ar[a4],ar[a2]))
						{	if (MOREEQ(ar[a4],ar[a3]))
							{	// 0 2 1 3 4 
								ipiv=a1;
							} else
							{	if (MORE(ar[a4],ar[a1]))
								{	// 0 2 1 4 3 
									ipiv=a1;
								} else
								{	// 0 2 4 1 3 
									ipiv=a4;
								}
							}
						} else
						{	if (MOREEQ(ar[a4],ar[a0]))
							{	// 0 4 2 1 3 
								ipiv=a2;
							} else
							{	// 4 0 2 1 3 
								ipiv=a2;
							}
						}
					} else
					{	if (MORE(ar[a4],ar[a2]))
						{	if (MOREEQ(ar[a4],ar[a1]))
							{	// 0 2 3 1 4 
								ipiv=a3;
							} else
							{	if (MORE(ar[a4],ar[a3]))
								{	// 0 2 3 4 1 
									ipiv=a3;
								} else
								{	// 0 2 4 3 1 
									ipiv=a4;
								}
							}
						} else
						{	if (MOREEQ(ar[a4],ar[a0]))
							{	// 0 4 2 3 1 
								ipiv=a2;
							} else
							{	// 4 0 2 3 1 
								ipiv=a2;
							}
						}
					}
				} else
				{	if (MOREEQ(ar[a3],ar[a0]))
					{	if (MOREEQ(ar[a4],ar[a3]))
						{	if (MOREEQ(ar[a4],ar[a1]))
							{	// 0 3 2 1 4 
								ipiv=a2;
							} else
							{	if (MOREEQ(ar[a4],ar[a2]))
								{	// 0 3 2 4 1 
									ipiv=a2;
								} else
								{	// 0 3 4 2 1 
									ipiv=a4;
								}
							}
						} else
						{	if (MOREEQ(ar[a4],ar[a0]))
							{	// 0 4 3 2 1 
								ipiv=a3;
							} else
							{	// 4 0 3 2 1 
								ipiv=a3;
							}
						}
					} else
					{	if (MOREEQ(ar[a4],ar[a0]))
						{	if (MOREEQ(ar[a4],ar[a1]))
							{	// 3 0 2 1 4 
								ipiv=a2;
							} else
							{	if (MOREEQ(ar[a4],ar[a2]))
								{	// 3 0 2 4 1 
									ipiv=a2;
								} else
								{	// 3 0 4 2 1 
									ipiv=a4;
								}
							}
						} else
						{	if (MORE(ar[a4],ar[a3]))
							{	// 3 4 0 2 1 
								ipiv=a0;
							} else
							{	// 4 3 0 2 1 
								ipiv=a0;
							}
						}
					}
				}
			}
		} else
		{	if (MOREEQ(ar[a3],ar[a0]))
			{	if (MOREEQ(ar[a3],ar[a1]))
				{	if (MOREEQ(ar[a4],ar[a0]))
					{	if (MOREEQ(ar[a4],ar[a3]))
						{	// 2 0 1 3 4 
							ipiv=a1;
						} else
						{	if (MORE(ar[a4],ar[a1]))
							{	// 2 0 1 4 3 
								ipiv=a1;
							} else
							{	// 2 0 4 1 3 
								ipiv=a4;
							}
						}
					} else
					{	if (MORE(ar[a4],ar[a2]))
						{	// 2 4 0 1 3 
							ipiv=a0;
						} else
						{	// 4 2 0 1 3 
							ipiv=a0;
						}
					}
				} else
				{	if (MOREEQ(ar[a4],ar[a0]))
					{	if (MOREEQ(ar[a4],ar[a1]))
						{	// 2 0 3 1 4 
							ipiv=a3;
						} else
						{	if (MORE(ar[a4],ar[a3]))
							{	// 2 0 3 4 1 
								ipiv=a3;
							} else
							{	// 2 0 4 3 1 
								ipiv=a4;
							}
						}
					} else
					{	if (MORE(ar[a4],ar[a2]))
						{	// 2 4 0 3 1 
							ipiv=a0;
						} else
						{	// 4 2 0 3 1 
							ipiv=a0;
						}
					}
				}
			} else
			{	if (MORE(ar[a3],ar[a2]))
				{	if (MOREEQ(ar[a4],ar[a3]))
					{	if (MOREEQ(ar[a4],ar[a1]))
						{	// 2 3 0 1 4 
							ipiv=a0;
						} else
						{	if (MORE(ar[a4],ar[a0]))
							{	// 2 3 0 4 1 
								ipiv=a0;
							} else
							{	// 2 3 4 0 1 
								ipiv=a4;
							}
						}
					} else
					{	if (MORE(ar[a4],ar[a2]))
						{	// 2 4 3 0 1 
							ipiv=a3;
						} else
						{	// 4 2 3 0 1 
							ipiv=a3;
						}
					}
				} else
				{	if (MOREEQ(ar[a4],ar[a2]))
					{	if (MOREEQ(ar[a4],ar[a1]))
						{	// 3 2 0 1 4 
							ipiv=a0;
						} else
						{	if (MOREEQ(ar[a4],ar[a0]))
							{	// 3 2 0 4 1 
								ipiv=a0;
							} else
							{	// 3 2 4 0 1 
								ipiv=a4;
							}
						}
					} else
					{	if (MORE(ar[a4],ar[a3]))
						{	// 3 4 2 0 1 
							ipiv=a2;
							return -1;
						} else
						{	// 4 3 2 0 1 
							ipiv=a2;
							return -1;
						}
					}
				}
			}
		}
	} else
	{	if (MOREEQ(ar[a2],ar[a1]))
		{	if (MOREEQ(ar[a2],ar[a0]))
			{	if (MOREEQ(ar[a3],ar[a0]))
				{	if (MOREEQ(ar[a3],ar[a2]))
					{	if (MOREEQ(ar[a4],ar[a0]))
						{	if (MOREEQ(ar[a4],ar[a3]))
							{	// 1 0 2 3 4 
								ipiv=a2;
								return 1;
							} else
							{	if (MOREEQ(ar[a4],ar[a2]))
								{	// 1 0 2 4 3 
									ipiv=a2;
									return 1;
								} else
								{	// 1 0 4 2 3 
									ipiv=a4;
								}
							}
						} else
						{	if (MORE(ar[a4],ar[a1]))
							{	// 1 4 0 2 3 
								ipiv=a0;
							} else
							{	// 4 1 0 2 3 
								ipiv=a0;
							}
						}
					} else
					{	if (MORE(ar[a4],ar[a0]))
						{	if (MOREEQ(ar[a4],ar[a2]))
							{	// 1 0 3 2 4 
								ipiv=a3;
							} else
							{	if (MORE(ar[a4],ar[a3]))
								{	// 1 0 3 4 2 
									ipiv=a3;
								} else
								{	// 1 0 4 3 2 
									ipiv=a4;
								}
							}
						} else
						{	if (MORE(ar[a4],ar[a1]))
							{	// 1 4 0 3 2 
								ipiv=a0;
							} else
							{	// 4 1 0 3 2 
								ipiv=a0;
							}
						}
					}
				} else
				{	if (MORE(ar[a3],ar[a1]))
					{	if (MOREEQ(ar[a4],ar[a3]))
						{	if (MOREEQ(ar[a4],ar[a2]))
							{	// 1 3 0 2 4 
								ipiv=a0;
							} else
							{	if (MOREEQ(ar[a4],ar[a0]))
								{	// 1 3 0 4 2 
									ipiv=a0;
								} else
								{	// 1 3 4 0 2 
									ipiv=a4;
								}
							}
						} else
						{	if (MORE(ar[a4],ar[a1]))
							{	// 1 4 3 0 2 
								ipiv=a3;
							} else
							{	// 4 1 3 0 2 
								ipiv=a3;
							}
						}
					} else
					{	if (MOREEQ(ar[a4],ar[a1]))
						{	if (MOREEQ(ar[a4],ar[a2]))
							{	// 3 1 0 2 4 
								ipiv=a0;
							} else
							{	if (MOREEQ(ar[a4],ar[a0]))
								{	// 3 1 0 4 2 
									ipiv=a0;
								} else
								{	// 3 1 4 0 2 
									ipiv=a4;
								}
							}
						} else
						{	if (MORE(ar[a4],ar[a3]))
							{	// 3 4 1 0 2 
								ipiv=a1;
							} else
							{	// 4 3 1 0 2 
								ipiv=a1;
							}
						}
					}
				}
			} else
			{	if (MOREEQ(ar[a3],ar[a2]))
				{	if (MORE(ar[a3],ar[a0]))
					{	if (MOREEQ(ar[a4],ar[a2]))
						{	if (MOREEQ(ar[a4],ar[a3]))
							{	// 1 2 0 3 4 
								ipiv=a0;
							} else
							{	if (MORE(ar[a4],ar[a0]))
								{	// 1 2 0 4 3 
									ipiv=a0;
								} else
								{	// 1 2 4 0 3 
									ipiv=a4;
								}
							}
						} else
						{	if (MORE(ar[a4],ar[a1]))
							{	// 1 4 2 0 3 
								ipiv=a2;
							} else
							{	// 4 1 2 0 3 
								ipiv=a2;
							}
						}
					} else
					{	if (MORE(ar[a4],ar[a2]))
						{	if (MOREEQ(ar[a4],ar[a0]))
							{	// 1 2 3 0 4 
								ipiv=a3;
							} else
							{	if (MORE(ar[a4],ar[a3]))
								{	// 1 2 3 4 0 
									ipiv=a3;
								} else
								{	// 1 2 4 3 0 
									ipiv=a4;
								}
							}
						} else
						{	if (MORE(ar[a4],ar[a1]))
							{	// 1 4 2 3 0 
								ipiv=a2;
							} else
							{	// 4 1 2 3 0 
								ipiv=a2;
							}
						}
					}
				} else
				{	if (MORE(ar[a3],ar[a1]))
					{	if (MOREEQ(ar[a4],ar[a3]))
						{	if (MOREEQ(ar[a4],ar[a0]))
							{	// 1 3 2 0 4 
								ipiv=a2;
							} else
							{	if (MOREEQ(ar[a4],ar[a2]))
								{	// 1 3 2 4 0 
									ipiv=a2;
								} else
								{	// 1 3 4 2 0 
									ipiv=a4;
								}
							}
						} else
						{	if (MORE(ar[a4],ar[a1]))
							{	// 1 4 3 2 0 
								ipiv=a3;
							} else
							{	// 4 1 3 2 0 
								ipiv=a3;
							}
						}
					} else
					{	if (MOREEQ(ar[a4],ar[a1]))
						{	if (MOREEQ(ar[a4],ar[a0]))
							{	// 3 1 2 0 4 
								ipiv=a2;
							} else
							{	if (MOREEQ(ar[a4],ar[a2]))
								{	// 3 1 2 4 0 
									ipiv=a2;
								} else
								{	// 3 1 4 2 0 
									ipiv=a4;
								}
							}
						} else
						{	if (MOREEQ(ar[a4],ar[a3]))
							{	// 3 4 1 2 0 
								ipiv=a1;
							} else
							{	// 4 3 1 2 0 
								ipiv=a1;
							}
						}
					}
				}
			}
		} else
		{	if (MOREEQ(ar[a3],ar[a1]))
			{	if (MOREEQ(ar[a3],ar[a0]))
				{	if (MOREEQ(ar[a4],ar[a1]))
					{	if (MOREEQ(ar[a4],ar[a3]))
						{	// 2 1 0 3 4 
							ipiv=a0;
						} else
						{	if (MORE(ar[a4],ar[a0]))
							{	// 2 1 0 4 3 
								ipiv=a0;
							} else
							{	// 2 1 4 0 3 
								ipiv=a4;
							}
						}
					} else
					{	if (MOREEQ(ar[a4],ar[a2]))
						{	// 2 4 1 0 3 
							ipiv=a1;
						} else
						{	// 4 2 1 0 3 
							ipiv=a1;
						}
					}
				} else
				{	if (MOREEQ(ar[a4],ar[a1]))
					{	if (MOREEQ(ar[a4],ar[a0]))
						{	// 2 1 3 0 4 
							ipiv=a3;
						} else
						{	if (MORE(ar[a4],ar[a3]))
							{	// 2 1 3 4 0 
								ipiv=a3;
							} else
							{	// 2 1 4 3 0 
								ipiv=a4;
							}
						}
					} else
					{	if (MOREEQ(ar[a4],ar[a2]))
						{	// 2 4 1 3 0 
							ipiv=a1;
						} else
						{	// 4 2 1 3 0 
							ipiv=a1;
						}
					}
				}
			} else
			{	if (MOREEQ(ar[a3],ar[a2]))
				{	if (MOREEQ(ar[a4],ar[a3]))
					{	if (MOREEQ(ar[a4],ar[a0]))
						{	// 2 3 1 0 4 
							ipiv=a1;
						} else
						{	if (MOREEQ(ar[a4],ar[a1]))
							{	// 2 3 1 4 0 
								ipiv=a1;
							} else
							{	// 2 3 4 1 0 
								ipiv=a4;
							}
						}
					} else
					{	if (MOREEQ(ar[a4],ar[a2]))
						{	// 2 4 3 1 0 
							ipiv=a3;
						} else
						{	// 4 2 3 1 0 
							ipiv=a3;
						}
					}
				} else
				{	if (MORE(ar[a4],ar[a2]))
					{	if (MOREEQ(ar[a4],ar[a0]))
						{	// 3 2 1 0 4 
							ipiv=a1;
						} else
						{	if (MORE(ar[a4],ar[a1]))
							{	// 3 2 1 4 0 
								ipiv=a1;
							} else
							{	// 3 2 4 1 0 
								ipiv=a4;
							}
						}
					} else
					{	if (MOREEQ(ar[a4],ar[a3]))
						{	// 3 4 2 1 0 
							ipiv=a2;
							return -1;
						} else
						{	// 4 3 2 1 0 
							ipiv=a2;
							return -1;
						}
					}
				}
			}
		}
	}
	return 0;
}



__declspec(thread) Elem *tar=NULL;
__declspec(thread) size_t prev_size=0;

inline void Sort::InitTripleStateQuickSortTar(size_t size)
{	// Initializes the temporary array used in the algorithm.
	// MUST be called before the first time _TripleStateQuickSort() is called.
	// And every time a larger array is sorted. This calling is done automatically by TripleStateQuickSort()
	//static int prev_sizepc=0;
	//int size = b-a+1;
	
	if (size>prev_size)
	{	free(tar);
		tar = (Elem *) malloc(((size>>1)+2)*sizeof(*tar));	
		prev_size=size;
	} else
	if (size<(prev_size>>2))
	{	tar = (Elem *) realloc(tar,((size>>1)+2)*sizeof(*tar));	
		prev_size=size;
	}
}

void Sort::FreeTripleStateQuickSortStorage()
{	free(tar); tar=NULL; prev_size=0;
	
}

void Sort::TripleStateQuickSort(Elem ar[],int a, int b _DEFINE_CBFUNC)
{	/* InitTripleStateQuickSortTar has to be called at least once. For latter calls, You might be able to ingnore calling InitTripleStateQuickSortTar
	If you understand how it works. It doesn't always reallocate temporary space. */
	InitTripleStateQuickSortTar(b-a+1);
	_TripleStateQuickSort(ar,a,b _CALL_CBFUNC);
}

inline void Sort::_TripleStateQuickSort(Elem ar[],int a, int b _DEFINE_CBFUNC)
{	int l,r,m,mr,ml,mid,i,j,n,k,k2;
	short use_temp;

	#ifdef RANDOM_SAMPLES
	// Internal Random Generator
	static unsigned int const r_a=16807,r_m=2147483648;			//r_c=0
	static unsigned int zgen = clock();							// Done only when program starts because of 'static'.
	double dran,dran2;
	#endif

	Elem p,temp,temp2;
	int lc,lc2;			// [L]ast [C]omparison result.

	#define LFEW_EQUALS(ml,mr) (mr-ml)<=(ml-l)>>2
	#define RFEW_EQUALS(ml,mr) (mr-ml)<=(r-mr)>>2

	#define MLEFT()													\
	{	do															\
		{	ml--; if (ml==l) goto exit2;							\
		} while (EQUAL_SLC(ar[ml],p,lc));					\
		m=ml; ml++;													\
		if (LFEW_EQUALS(ml,mr))										\
		{	CONSERVE_INVALID(j); goto m_scan3L_2;					\
		}															\
		else														\
		{	goto m_scan2L_2;										\
		}															\
	}

	#define MRIGHT()												\
	{	do															\
		{	mr++; if (mr==r) goto exit2;							\
		} while (EQUAL_SLC(ar[mr],p,lc));					\
		m=mr; mr--;													\
		if (RFEW_EQUALS(ml,mr))										\
		{	CONSERVE_INVALID(j); goto m_scan3R_2;					\
		}															\
		else														\
		{	goto m_scan2R_2;										\
		}															\
	}

	#define MLEFT()													\
	{	do															\
		{	ml--; if (ml==l) goto exit2;							\
		} while (EQUAL_SLC(ar[ml],p,lc));					\
		m=ml; ml++;													\
		if (LFEW_EQUALS(ml,mr))										\
		{	CONSERVE_INVALID(j); goto m_scan3L_2;					\
		}															\
		else														\
		{	goto m_scan2L_2;										\
		}															\
	}

	#define MRIGHT_NOSCAN()											\
	{	m=mr; mr--;													\
		if (RFEW_EQUALS(ml,mr))										\
		{	CONSERVE_INVALID(j); goto m_scan3R_2;					\
		}															\
		else														\
		{	goto m_scan2R_2;										\
		}															\
	}

	#define MLEFT_NOSCAN()											\
	{	m=ml; ml++;													\
		if (LFEW_EQUALS(ml,mr))										\
		{	CONSERVE_INVALID(j); goto m_scan3L_2;					\
		}															\
		else														\
		{	goto m_scan2L_2;										\
		}															\
	}

	#ifdef CONSERVE_MORE_COPS
	#define MLEFT_CHECKM()											\
	{	/*if (m==ml) if (ml!=mr) { ar[ml]=ar[mr]; }*/		\
		j=ml;														\
		do															\
		{	ml--; if (ml==l) { if (m==j) if (j!=mr) { ar[j]=ar[mr]; } goto exit2; }	\
		} while (EQUAL_SLC(ar[ml],p,lc));					\
																	\
		ml++;														\
		if (LFEW_EQUALS(ml,mr))										\
		{	if (m==j) { if (j!=mr) { tar[i++]=ar[mr]; } } else CONSERVE_INVALID(j);	\
			m=ml-1; goto m_scan3L_2;								\
		}															\
		else														\
		{	if (m==j) if (j!=mr) { ar[j]=ar[mr]; }			\
			m=ml-1; goto m_scan2L_2;								\
		}															\
	}

	#define MRIGHT_CHECKM()											\
	{	/*if (m==ml) if (ml!=mr) { ar[ml]=ar[mr]; }*/		\
		j=mr;														\
		do															\
		{	mr++; if (mr==r) { if (m==j) if (j!=ml) { ar[j]=ar[ml]; } goto exit2; }	\
		} while (EQUAL_SLC(ar[mr],p,lc));					\
																	\
		mr--; 														\
		if (RFEW_EQUALS(ml,mr))										\
		{	if (m==j) { if (j!=ml) { tar[i++]=ar[ml]; } } else CONSERVE_INVALID(j);	\
			m=mr+1; goto m_scan3R_2;								\
		}															\
		else														\
		{	if (m==j) if (j!=ml) { ar[j]=ar[ml]; }			\
			m=mr+1; goto m_scan2R_2;								\
		}															\
	}

	#else

	#define MLEFT_CHECKM()											\
	{	if (m==ml) if (ml!=mr) { ar[ml]=ar[mr]; }			\
		MLEFT()														\
	}

	#define MRIGHT_CHECKM()											\
	{	if (m==mr) if (mr!=ml) { ar[mr]=ar[ml]; }			\
		MRIGHT()													\
	}
	#endif

	#ifdef SWAP_P_TO_MID
		#define COPYPBACK() 
	#else
		#define COPYPBACK() if (a2!=mid) { ar[mid]=p; }
	#endif

	if (b-a<TINY_THRESH3) { l=b; goto insertion_sort; }
	i=0; use_temp=1;

	#ifdef RANDOM_SAMPLES
	// Generate internal random number
	zgen=(r_a*zgen) % r_m;
	//double dran = 2+((double) (zgen-r_m/2))/r_m;
	dran = 0.5+(double) zgen/r_m;
	dran2 = 1 + dran;
	#endif

	goto skip_pop;

	do
	{	b=POP(stack1); a=POP(stack1);
		
		skip_pop:
		n=b-a+1;
		int a0,a1,a2,a3;

		// Agressive Median System
		int gap,gap2,gap2s,cl,cr;	//(gap = n/(5-1))
		mid = (b+a)>>1;
		short flag;

		// Test for small sets first because these are the most common in the sorting tree
		// Then mid size, then large sized sets.
		#ifdef MEDOF3_SMALL
		// If n is small use median of three
		if (n<=SMALL_THRESH)
		{	l=a+1; r=b-1;
			// When n<200, Median of 3 generally gives better comparisons but a little more swaps than median of 5
			// Median of 5 is a little more robust against getting into O(n^2)
			flag = MedianOf3S(ar,a,mid,b,p _CALL_CBFUNC);
			switch (flag) 
			{	
				case -1: 
				#ifndef SWAP_P_TO_MID
					a2=mid;
				#endif
				goto possibly_reversed; 

				case 0: goto main_algo;

				default:
				#ifndef SWAP_P_TO_MID
					a2=mid;
				#endif
				goto possibly_sorted;
			}
		} else
		#endif
		if (n<=MID_THRESH)
		{	// Small section
			#ifdef RANDOM_SAMPLES
				gap=(n>>2)*dran;
			#else
				gap=(n>>2);
			#endif
			l=a+1; r=b-1;
			// When n<200, Median of 3 generally gives better comparisons but a little more swaps than median of 5
			// Median of 5 is a little more robust against getting into O(n^2)
			
			flag = MedianOf5S(ar,a,a + gap,mid,b - gap,b,p _CALL_CBFUNC);
			switch (flag)
			{	
				case -1: 
				#ifndef SWAP_P_TO_MID
					a2=mid;
				#endif
				goto possibly_reversed; 

				case 0: goto main_algo;

				default:
				#ifndef SWAP_P_TO_MID
					a2=mid;
				#endif
				possibly_sorted:
				// Possibly Sorted Handler
				while (LESS_SLC(ar[l],p,lc2))
				{	l++; 
					if (l==mid)
					{	while (MORE_SLC(ar[r],p,lc2))
						{	r--; if (r==mid) { r++; l--; COPYPBACK(); goto bypass; }
						}
						if (lc2==0)
						{	mr=mid+1; if (mr==r) { r++; l--; COPYPBACK(); goto bypass; }
							while (EQUAL_SLC(ar[mr],p,lc))
							{	mr++; if (mr==r) { r++; l--; COPYPBACK(); goto bypass; }
							}
							temp=ar[r]; ml=mid; RESETFL(); MRIGHT_NOSCAN(); //**
						} else
						{	temp=ar[r]; 
							ml=mr=mid;
							//if (mr==r) goto exit2;
							RESETFL();
							MRIGHT();
						}
					}
				}
				if (lc2==0) 
				{	while (MORE_SLC(ar[r],p,lc))
					{	r--; 
						if (r==mid)
						{	ml=mid-1; if (ml==l) { r++; l--; COPYPBACK(); goto bypass; }
							while (EQUAL_SLC(ar[ml],p,lc))
							{	ml--; if (ml==l) { r++; l--; COPYPBACK(); goto bypass; }
							}
							temp=ar[l]; mr=mid; RESETFL(); MLEFT_NOSCAN();
						}
					}

					if (lc==0)
					{	ml=mid; 
						do
						{	ml--;
							if (ml==l)
							{	mr=mid+1; if (mr==r) { r++; l--; COPYPBACK(); goto bypass; }
								while (EQUAL_SLC(ar[mr],p,lc))
								{	mr++; if (mr==r) { r++; l--; COPYPBACK(); goto bypass; }
								}
								temp=ar[l]; ar[mid]=ar[r]; RESETFL(); MRIGHT_NOSCAN();
							}
						} while (EQUAL_SLC(ar[ml],p,lc));
						mr=m=mid; temp=ar[l]; ar[mid]=ar[r]; RESETFL(); goto ml_scan1_2;	//** mr=m=mid; try without m
					} else
					{	temp=ar[l]; m=mr=ml=mid; RESETFL(); goto r_scan1_3;	//** m=mr=ml=mid; try without m
					}
				}

				// Ooops it's not sorted. assumption is wrong, we'll mitigate.
				// Check for special cases.				
				while (MORE_SLC(ar[r],p,lc2))
				{	r--; if (r==mid) { use_temp=0; CONSERVE_INVALID(j); m=mid; RESETFL(); goto checkp11; }			// Actualy m=mr=ml=mid here, so it doesn't matter what you use.
				}
				temp=ar[r]; m=mr=ml=mid; RESETFL(); goto l_scan1_2;

			}
		}
		else if (n<=LARGE_THRESH)
		{	// Mid sized section
			#ifdef RANDOM_SAMPLES
				gap = (n>>3)*dran;
				gap2 = (n>>3)*dran2;
			#else
				gap2 = (gap=(n>>3))<<1;
			#endif
			flag=MedianOf3NS(ar,MedOfThreeNoSort(ar,a,a+gap,a+gap2 _CALL_CBFUNC)	\
							,MedOfThreeNoSort(ar,mid-gap,mid,mid+gap _CALL_CBFUNC)	\
							,MedOfThreeNoSort(ar,b-gap2,b-gap,b _CALL_CBFUNC)	\
							,a2 _CALL_CBFUNC);
			p=ar[a2]; l=a; r=b;

			#ifdef SWAP_P_TO_MID
				if (a2!=mid) SWAP_AR(ar,a2,mid,temp2);	// Works without it!!, but it shouldn't check why
			#else
				if (a2!=mid) { ar[a2]=ar[mid]; }
			#endif
			switch (flag)
			{	case 1: goto possibly_sorted; 
				case 0: goto main_algo;
				default:
				possibly_reversed:
				// Possibly Reversed
				// You could right this in a more readable way, but this is a bit more optimized for speed. looks weired, I know.
				cl=1; cr=1;	// counters. cl for ar[l]<p and cr for ar[r]>p
				for (;;)
				{	if (MORE_SLC(ar[l],p,lc2))
					{
						checkp2:
						if (LESS_SLC(ar[r],p,lc))
						{	SWAP_AR(ar,l,r,temp2);
							if (l+2==r) { COPYPBACK(); goto bypass; }
							l++; r--;
							if (l==mid) { mr=mid+1; goto checkp7; }
							else if (r==mid)
							{	ml=mid;
								do
								{	ml--; 
									if (ml==l)
									{	// Special case short cut.
										if (MORE_SLC(ar[l],p,lc))
										{	ar[mid]=ar[l]; ar[l]=p; l--;
											goto bypass; // no need for COPYPBACK
										}
										else if (lc==0) { l--; r++; }
										else r++;
										COPYPBACK(); goto bypass;
									}
								} while (EQUAL_SLC(ar[ml],p,lc));
								MORE_SLC_NOOP(ar[l],p,lc2);
								temp=ar[l]; mr=mid; RESETRFL(); MLEFT_NOSCAN();
							}
						}
						else if (lc>0) 
						{	r--;
							if (r==mid)
							{	ml=mid;
								do
								{	ml--; 
									if (ml==l)
									{	// Special case short cut.
										ar[mid]=ar[l]; ar[l]=p; l--; /*COPYPBACK();*/ goto bypass; // no COPYPBACK because of special case handling
									}
								} while (EQUAL_SLC(ar[ml],p,lc));
								temp=ar[l]; mr=mid; RESETRFL(); MLEFT_NOSCAN();
							}
							if (cr>REVERSE_TOLERANCE)
							{	temp=ar[l]; m=mr=ml=mid; RESETRFL(); goto r_scan1;
							}
							else { cr++; goto checkp2; }
						}
						else	//lc==0
						{	mr=mid;
							do
							{	mr++;
								if (mr==r)
								{	ml=mid;
									do
									{	ml--;
										if (ml==l)
										{	SWAP_AR(ar,l,r,temp2); l--; COPYPBACK(); goto bypass;
										}
									} while (EQUAL_SLC(ar[ml],p,lc));
									ar[mid]=ar[r]; temp=ar[l]; 
									RESETRFL();
									MLEFT_NOSCAN();
								}
							} while (EQUAL_SLC(ar[mr],p,lc));
							temp=ar[l]; ar[mid]=ar[r]; ml=mid; RESETRFL(); goto mr_scan1_2;
						}
					}
					else if (lc2<0) 
					{	l++; 
						if (l==mid)
						{	mr=mid;
							do
							{	mr++; 
								checkp7:
								if (mr==r)
								{	// Special case short cut.
									if (LESS_SLC(ar[r],p,lc))
									{	ar[l]=ar[r]; ar[r]=p; r++; goto bypass; // no need for COPYPBACK 
									}
									else if (lc==0) { l--; r++; }
									else l--;
									COPYPBACK(); goto bypass; 
								}
							} while (EQUAL_SLC(ar[mr],p,lc));
							while (MORE_SLC(ar[r],p,lc2))
							{	r--;
								if (mr==r)
								{	if (lc<0) { ar[l]=ar[r]; ar[r]=p; r++; goto bypass; } //no need for COPYPBACK
									else l--;
									COPYPBACK(); goto bypass;
								}
							}
							temp=ar[r]; ml=mid; RESETRFL(); MRIGHT_NOSCAN();
						}
						if (cl>REVERSE_TOLERANCE) { RESETRFL(); goto main_algo; }
						else { cl++; continue; } 
					}
					else	//lc2==0
					{	ml=mid; 
						do
						{	ml--; 
							if (ml==l)
							{	mr=mid;
								do
								{	mr++;
									if (mr==r) 
									{	if (LESS_SLC(ar[r],p,lc)) { SWAP_AR(ar,r,l,temp2); r++; }
										else { l--; if (lc==0) r++; }
										COPYPBACK(); goto bypass;
									}
								} while (EQUAL_SLC(ar[mr],p,lc));
								// Try scanning from right before reverting to normal method
								while (MORE_SLC(ar[r],p,lc2))
								{	r--; 
									if (mr==r)
									{	if (lc<0) { SWAP_AR(ar,mr,ml,temp2); r++; }
										else l--;
										COPYPBACK(); goto bypass;
									}
								}
								temp=ar[r]; ar[mid]=ar[ml]; RESETRFL(); MRIGHT_NOSCAN();
							}
						} while (EQUAL_SLC(ar[ml],p,lc2));
						ar[mid]=ar[l]; temp=ar[ml]; 		//** try scanning from right instead (may be better)
						m=ml; mr=mid; RESETRFL(); goto r_scan1;
					}
					
				} // for(;;)
			}
		}
		else
		{	// Large section
			// This will reduce comparisons and swaps but has some extra overhead. 
			// Lower LARGE_THRESH when the comparison function is CPU intensive.
			gap=n>>2;
			a1 = a + gap;
			a3 = b - gap;
			
			#ifdef RANDOM_SAMPLES
				gap2s = (gap2=n/14)*dran2;
				gap2*=dran;
			#else
				gap2s = (gap2=n/14)<<1;			//n/((5-1)*3+)
			#endif
			flag=MedianOf5NS(ar,MedOfThreeNoSort(ar,a,a+gap2,a+gap2s _CALL_CBFUNC)	\
							,MedOfThreeNoSort(ar,a1,a1+gap2,a1+gap2s _CALL_CBFUNC)	\
							,MedOfThreeNoSort(ar,mid-gap2,mid,mid+gap2 _CALL_CBFUNC)	\
							,MedOfThreeNoSort(ar,a3-gap2s,a3-gap2,a3 _CALL_CBFUNC)	\
							,MedOfThreeNoSort(ar,b-gap2s,b-gap2,b _CALL_CBFUNC)	\
							,a2 _CALL_CBFUNC);
			p=ar[a2]; l=a; r=b;

			#ifdef SWAP_P_TO_MID
				if (a2!=mid) SWAP_AR(ar,a2,mid,temp2);
			#else
				if (a2!=mid) { ar[a2]=ar[mid]; }
			#endif
			switch (flag)
			{
				case 1: 
				goto possibly_sorted;


				case -1: 
				goto possibly_reversed;
				

				default: ;
				//goto main_algo;		// no need
			}
		}

		// Normal case
		main_algo:

		// Pre-scan from the right
		while (MORE_SLC(ar[r],p,lc2))
		{	r--; 
			if (r==mid) 
			{	use_temp=0; CONSERVE_INVALID(j); m=mid; //goto l_scan3L;  // Actualy m=mr=ml=mid here, so it doesn't matter what you use.
				
				// Special case, there might be no equals to pivot.
				for (;;)
				{	while (LESS_SLC(ar[l],p,lc))
					{	l++; if (m==l) goto exit3L;
					}
					if (lc>0)
					{	
						checkp11:
						ar[r]=ar[l]; r--; 
						m--; if (m==l) goto exit3L;
					} else
					{	tar[i++]=ar[l]; m--;
						if (m==l) goto exit3L; else { ml=mid; goto m_scan3L; }
					} 
					
					while (MORE_SLC(ar[m],p,lc))
					{	m--; if (m==l) { r=m; goto exit3L; }
					}
					r=m;
					if (lc<0)
					{	ar[l]=ar[m]; l++; if (m==l) goto exit3L;
					} else
					{	tar[i++]=ar[m]; m--; if (m==l) goto exit3L; else { ml=mid; goto m_scan3L; }
					}
				}
			}			
		}

		#ifdef RL_PRESCAN
		// Pre-scan from the left
		while (LESS_SLC(ar[l],p,lc))
		{	l++; 
			if (l==mid) 
			{ use_temp=0; CONSERVE_INVALID(j); m=mid; //goto r_scan3R; // Actualy m=mr=ml=mid here, so it doesn't matter what you use.
			
				// Special case, there might be no equals to pivot.
				for (;;)
				{	while (MORE_SLC(ar[r],p,lc))
					{	r--; if (m==r) goto exit3R;
					}
					if (lc<0)
					{	ar[l]=ar[r]; l++; 
						m++; if (m==r) goto exit3R;
					} else
					{	tar[i++]=ar[r]; m++;
						if (m==r) goto exit3R; else { mr=mid; goto m_scan3R; }
					} 
					
					while (LESS_SLC(ar[m],p,lc))
					{	m++; if (m==r) { l=m; goto exit3R; }
					}
					l=m;
					if (lc>0)
					{	ar[r]=ar[m]; r--; if (m==r) goto exit3R;
					} else
					{	tar[i++]=ar[m]; m++; if (m==r) goto exit3R; else { mr=mid; goto m_scan3R; }
					}
				}
			}			
		}
		#endif

		// State 1. m moving right and left, both mid sides open
		state1:
		
		m=mr=ml=mid; //temp=ar[r];
		// ml and mr always point to the ends of the equals range that we discovered so far (inclusive). So they 
		// NOT point to an untested element unless we are about to test it in the next step.
		temp=ar[r]; 
		#ifdef RL_PRESCAN
		goto l_scan1_4;
		#endif

		// Pre-scan from the left
		l_scan1:
		while (LESS_SLC(ar[l],p,lc))
		{	l++; if (l==ml) MRIGHT_CHECKM();
		}
		l_scan1_4:
		if (lc==0) 
		{	
			l_scan1_3:
			ar[m]=ar[l]; 
			l_scan1_5:
			if (r-mr>ml-l)		// choose largest gap
			{	mr++; /*if (mr==r) { MLEFT() } else*/ goto mr_scan1; // Checking IS not neccessary here
			} else
			{	ml--; if (ml==l) { MRIGHT() } else goto ml_scan1;
			}
		} else 
		{	
			l_scan1_2:
			ar[r]=ar[l]; r--;
			// You could try choose largest gap here also
			if (mr==r) MLEFT_CHECKM() else goto r_scan1; 
		}

		mr_scan1:
		while (EQUAL_SLC(ar[mr],p,lc))
		{	mr++; if (mr==r) MLEFT();
		}
		mr_scan1_2:
		if (lc<0) 
		{	ar[l]=ar[mr]; l++;
			if (ml==l) 
			{	ar[mr]=ar[ml]; MRIGHT();
			} else { m=mr; goto l_scan1; } 
		} else 
		{	ar[r]=ar[mr]; r--;
			if (mr==r) MLEFT() else { m=mr; goto r_scan1; }
		}

		ml_scan1:
		while (EQUAL_SLC(ar[ml],p,lc))
		{	ml--; if (ml==l) MRIGHT();
		}
		ml_scan1_2:
		if (lc<0) 
		{	ar[l]=ar[ml]; l++; 
			if (l==ml) MRIGHT() else { m=ml; goto l_scan1; }
		} else 
		{	ar[r]=ar[ml]; r--; 
			if (mr==r) 
			{	ar[ml]=ar[mr]; MLEFT();
			} else { m=ml; goto r_scan1; }
		}

		r_scan1:
		while (MORE_SLC(ar[r],p,lc))
		{	r--; 
			if (mr==r) MLEFT_CHECKM();
		}
		r_scan1_2:
		if (lc==0)
		{	ar[m]=ar[r]; 
			r_scan1_4:
			if (r-mr>ml-l)		// choose largest gap
			{	mr++; /*if (mr==r) MLEFT() else*/ goto mr_scan1; // Checking IS not neccessary here
			} else
			{	ml--; if (ml==l) MRIGHT() else goto ml_scan1;
			}
		} else 
		{	
			r_scan1_3:
			ar[l]=ar[r]; l++; 

			// You could try choose largest gap here also
			if (ml==l) MRIGHT_CHECKM() else goto l_scan1; 
		}

		// -----------------
		// State 2L: m moving left, right side closed
		//state2L:
		m_scan2L:
		while (EQUAL_SLC(ar[m],p,lc))
		{	m--; if (m==l) goto exit2;
		}
		m_scan2L_2:
		if (lc<0) 
		{	ar[l]=ar[m]; l++; 
			if (m==l) goto exit2; else goto l_scan2L; 
		} else 
		{	ar[r]=ar[m]; r--; ar[m]=ar[r]; m--; 
			if (m==l) goto exit2; else goto m_scan2L;
		}
		l_scan2L:
		while (LESS_SLC(ar[l],p,lc))
		{	l++; if (m==l) goto exit2;
		}
		if (lc==0) 
		{	ar[m]=ar[l]; m--;
			if (m==l) goto exit2; else goto m_scan2L;
		} else 
		{	ar[r]=ar[l]; r--; ar[m]=ar[r]; m--; 
			if (m==l) goto exit2; else goto m_scan2L;
		}

		// State 3L: m moving left, right side closed
		// Second method, when there are too little equal values
		//state3L:
		m_scan3L:
		while (EQUAL_SLC(ar[m],p,lc))
		{	
			m_scan3L_3:
			tar[i++]=ar[m]; m--; if (m==l) goto exit3L;
		}
		m_scan3L_2:
		if (lc<0) 
		{	
			m_scan3L_4:
			ar[l]=ar[m]; l++;
			if (m==l) goto exit3L; else goto l_scan3L;
		} else 
		{	// Check for a run of elements >p
			k=m;
			do
			{	m--; if (m==l) break;
			} while (MORE_SLC(ar[m],p,lc));
				
			k2=m+1;
			// Copy the run to the gap between r and m
			if (k-m<r-k)
				do
				{	ar[r]=ar[k2]; r--;
					if (r>=ml CONSERVE_CHECK(r,j)) { tar[i++]=ar[r]; }
				} while ((++k2)<=k);
			else
				for (;;)
				{	ar[r]=ar[k2]; r--;
					if (r>=ml CONSERVE_CHECK(r,j)) { tar[i++]=ar[r]; }
					else if (r<=k) { r=k2; break; }
					k2++;
				}
			if (m==l) goto exit3L; else if (lc==0) goto m_scan3L_3; else goto m_scan3L_4;
			
		}
		l_scan3L:
		while (LESS_SLC(ar[l],p,lc))
		{	l++; if (m==l) goto exit3L;
		}
		if (lc==0)
		{	
			l_scan3L_3:
			tar[i++]=ar[l]; m--;
			if (m==l) goto exit3L; else goto m_scan3L;
		} else 
		{	
			l_scan3L_2:
			ar[r]=ar[l]; r--; 
			// If ar[r] is not empty, make it
			if (r>=ml CONSERVE_CHECK(r,j)) { tar[i++]=ar[r]; }
			m--;
			if (m==l) goto exit3L; else goto m_scan3L;
		}

		// ------------------
		// State 2R. m moving right, left side closed
		//state2R:
		m_scan2R:
		while (EQUAL_SLC(ar[m],p,lc))
		{	m++; if (m==r) goto exit2;
		}
		m_scan2R_2:
		if (lc<0) 
		{	ar[l]=ar[m]; l++; ar[m]=ar[l]; m++; 
			if (m==r) goto exit2; else goto m_scan2R;
		} else 
		{	ar[r]=ar[m]; r--; 
			if (m==r) goto exit2; else goto r_scan2R;
		}
		r_scan2R:
		while (MORE_SLC(ar[r],p,lc))
		{	r--; if (m==r) goto exit2;
		}
		if (lc==0)
		{	ar[m]=ar[r]; m++; 
			if (m==r) goto exit2; else goto m_scan2R;
		} else 
		{	ar[l]=ar[r]; l++; ar[m]=ar[l]; m++; 
			if (m==r) goto exit2; else goto m_scan2R;
		}

		// Second method, when there are too little equal values
		// State 3R: m moving right, left side closed
		//state3R:
		m_scan3R:
		while (EQUAL_SLC(ar[m],p,lc))
		{	
			m_scan3R_3:
			tar[i++]=ar[m]; m++; if (m==r) goto exit3R;
		}
		m_scan3R_2:
		if (lc>0) 
		{	
			m_scan3R_4:
			ar[r]=ar[m]; r--; 
			if (m==r) goto exit3R; else goto r_scan3R;
		} else 
		{	k=m;
			do
			{	m++; if (m==r) break;
			} while (LESS_SLC(ar[m],p,lc));
			
			k2=m-1;
			// Copy only last elements
			if (m-k<k-l)
				do
				{	ar[l]=ar[k2]; l++;
					if (l<=mr CONSERVE_CHECK(l,j)) { tar[i++]=ar[l]; }
				} while ((--k2)>=k);
			else
				for (;;)
				{	ar[l]=ar[k2]; l++;
					if (l<=mr CONSERVE_CHECK(l,j)) { tar[i++]=ar[l]; }
					else if (l>=k) { l=k2; break; }
					k2--;
				}
			if (m==r) goto exit3R; else if (lc==0) goto m_scan3R_3; else goto m_scan3R_4;
		}

		r_scan3R:
		while (MORE_SLC(ar[r],p,lc))
		{	r--; if (m==r) goto exit3R;
		}
		if (lc==0)
		{	tar[i++]=ar[r]; m++;
			if (m==r) goto exit3R; else goto m_scan3R;
		} else 
		{	ar[l]=ar[r]; l++; 
			if (l<=mr CONSERVE_CHECK(l,j)) { tar[i++]=ar[l]; }
			m++;
			if (m==r) goto exit3R; else goto m_scan3R;
		}

		exit3L:
		if (use_temp)
		{	
			#ifdef CONSERVE_MORE_COPS
			if (j!=INVALID) 
				if (j<r) { i--; ar[j]=tar[i]; } 
			#endif
			while(i>0)
			{	i--; m++; ar[m]=tar[i]; 
			}
			goto exit2;
		} else
		{	// Special case when temp is not used.
			use_temp=1;
			while(i>0)
			{	i--; m++; ar[m]=tar[i]; 
			}
			r++;
			ar[l]=p; l--; goto bypass;
		}

		exit3R:
		#ifdef RL_PRESCAN
		if (use_temp)
		{	
		#endif
			#ifdef CONSERVE_MORE_COPS
			if (j!=INVALID) 
				if (j>l) { i--; ar[j]=tar[i]; } 
			#endif
			while (i>0)
			{	i--; m--; ar[m]=tar[i]; 
			}
		#ifdef RL_PRESCAN
		} else
		{	// Special case when temp is not used.
			use_temp=1;
			while(i>0)
			{	i--; m--; ar[m]=tar[i]; 
			}
			l--;
			ar[r]=p; r++; goto bypass;
		}
		#endif

		exit2:
		if (lc2>=0)
		{	ar[r]=temp; ar[l]=p; l--; 	//** if (l!=mid);
			if (lc2==0) r++;
		} else
		{	ar[l]=temp; ar[r]=p; r++; 
		}

		bypass:
		
		// In this algo, r and l do NOT cross each other. Thats why you l inplace of r and vise versa
		recurse:
		#ifdef DUAL_TINY_THRESH
		if (b-r>TINY_THRESH || (NOT_SORT_REV() && b-r>TINY_THRESH2))
		#else
		if (b-r>TINY_THRESH)
		#endif
		{	PUSH(stack1,r); PUSH(stack1,b);
		} else
		#ifdef USE_SortN
		switch (b-r)
		{	//case 5: Sort_6(ar+r _CALL_CBFUNC); break;
			//case 4: Sort_5(ar+r _CALL_CBFUNC); break;
			case 3: Sort_4(ar+r _CALL_CBFUNC); break;
			case 2: Sort_3(ar+r _CALL_CBFUNC); break;
			case 1: Sort_2(ar+r _CALL_CBFUNC); break;
			case 0: break;
			default:
		#else
		{
		#endif
			// Insertion sort
			for (k=r+1; k<=b; k++)
			{	if (LESS(ar[k],ar[k-1]))
				{	j=k; temp=ar[j];
					ar[j] = ar[j-1]; 
					if (--j>r)
						while (LESS(temp,ar[j-1]))
						{	ar[j] = ar[j-1]; 
							if (--j==r) break;
						}
					ar[j] = temp; 
				}
			}
		}

		// No need to push the other ends because we are going to imdiatly pop it in the next cycle. just place it in a and b. Even no need to change a in this case!!
		#ifdef DUAL_TINY_THRESH
		if (l-a>TINY_THRESH || (NOT_SORT_REV() && l-a>TINY_THRESH2))
		#else
		if (l-a>TINY_THRESH)
		#endif
		{	b=l; goto skip_pop;
		} else
		insertion_sort:
		#ifdef USE_SortN
		switch (l-a)
		{	//case 5: Sort_6(ar+a _CALL_CBFUNC); break;
			//case 4: Sort_5(ar+a _CALL_CBFUNC); break;
			case 3: Sort_4(ar+a _CALL_CBFUNC); break;
			case 2: Sort_3(ar+a _CALL_CBFUNC); break;
			case 1: Sort_2(ar+a _CALL_CBFUNC); break;
			case 0: break;
			default:
		#else
		{
		#endif
			// Insertion sort
			for (k=a+1; k<=l; k++)
			{	if (LESS(ar[k],ar[k-1]))
				{	j=k; temp=ar[j];
					ar[j] = ar[j-1]; 
					if (--j>a)
						while (LESS(temp,ar[j-1]))
						{	ar[j] = ar[j-1]; 
							if (--j==a) break;
						}
					ar[j] = temp; 
				}
			}
		}	
	} while (NOTEMPTY(stack1)); 
}

