#pragma once

#include "Elem.h"

#define INVALID -10000
#define STACK_SIZE 200

#define CALL_BACK

#undef USE_SortN						// Uses dedicated sorting functions for tiny array sizes of 2,3 and 4 instead of insertion sort. Particularly usefull when you have tiny millions of tiny arrays of sizes 4,3. Otherwise, it seams to add overhead.

// Triple State Quicksort
#define TINY_THRESH		15				// 15 works good, specially for almost sorted or almost reversed data.
#define TINY_THRESH2	9				// Insertion sort is generaly good for sorted and -reversed- cases. but not really good for random cases
#define TINY_THRESH3	7
#undef SWAP_P_TO_MID					// Makes sure pivot is at mid from the beggining. This will take just a little bit more swaps to do, but will relieve a little overhead.
#define MEDOF3_SMALL					// Define to use median of 3 instead of median of five for small arrays. Tiny arrays are handeled using insertion sort
#undef RANDOM_SAMPLES					// Randomizes the gaps used to locate internal samples for pivoting Triple State QuickSort. This will help preventing n^2 case.
#undef RL_PRESCAN						// Adds Pre-scanning from left. So Triple State will scan from both right and left. 
#undef CONSERVE_MORE_COPS				// Can slightly conserve more swaps (copies) using a little more complex trick at the expense of some overhead. No Comparisons hit!!.
#define SMALL_THRESH		70			// Seams any thing between 40 and 100 is good (60)
#define MID_THRESH			170			// 150 to 180 looks good (160)
#define LARGE_THRESH		1000		// empirical is (300)
#define REVERSE_TOLERANCE	9
#undef DUAL_TINY_THRESH					// When array is NOT sorted (nor reversed) it will use a second better threshold for insertion sort

#ifdef DUAL_TINY_THRESH
	#undef USE_REVERSE
	#define RESETFL() flag=0
	#ifdef USE_REVERSE
		#define RESETRFL() flag=0
		#define NOT_SORT_REV() flag==0
	#else
		#define RESETRFL()
		#define NOT_SORT_REV() flag!=1
	#endif
#else
	#undef USE_REVERSE
	#define RESETFL()
	#define RESETRFL()
	#define NOT_SORT_REV()
#endif

#ifdef CONSERVE_MORE_COPS
	#define CONSERVE_INVALID_PC(j) j=NULL
	#define CONSERVE_INVALID(j) j=INVALID
	#define CONSERVE_CHECK(v,j) && v!=j
#else
	#define CONSERVE_INVALID_PC(j)
	#define CONSERVE_INVALID(j)
	#define CONSERVE_CHECK(v,j)
#endif

/*
CALL_BACK means that the comparison function is provided by the user after compilation. This is the usual case
Undefining CALL_BACK will cause usage of direct copmarison operators < > <= >= ==, this will lead to great improvement in run time.
but prevent the user from doing his own comparison function, he has to recompile the code to do so.
*/
#ifdef CALL_BACK

	#define LESS(a1,a2) cmp(&a1,&a2)<0
	#define LESSEQ(a1,a2) cmp(&a1,&a2)<=0
	#define MORE(a1,a2) cmp(&a1,&a2)>0
	#define MOREEQ(a1,a2) cmp(&a1,&a2)>=0
	#define EQUAL(a1,a2) cmp(&a1,&a2)==0
	// LC stands for Last Comparison. SLC means STORE Last Comparison
	#define LESS_SLC(a1,a2,lc) (lc = cmp(&a1,&a2))<0
	#define LESSEQ_SLC(a1,a2,lc) (lc = cmp(&a1,&a2))<=0
	#define MORE_SLC(a1,a2,lc) (lc = cmp(&a1,&a2))>0
	#define MOREEQ_SLC(a1,a2,lc) (lc = cmp(&a1,&a2))>=0
	#define EQUAL_SLC(a1,a2,lc) (lc = cmp(&a1,&a2))==0
	#define MORE_SLC_NOOP(a1,a2,lc) lc = cmp(&a1,&a2)

	// define Call Back (CB) FUnction in function definitions (a little not readable but saves me allot of typing)
	#define _CALL_CBFUNC	,cmp
	#define _CALL_CBFUNC_COMP ,compare
	#define _DEFINE_CBFUNC	,int cmp(const void *elem1,const void *elem2)
#else
	
	#define LESS(a1,a2) a1<a2
	#define LESSEQ(a1,a2) a1<=a2
	#define MORE(a1,a2) a1>a2
	#define MOREEQ(a1,a2) a1>=a2
	#define EQUAL(a1,a2) a1==a2
	// LC stands for Last Comparison. SLC means STORE Last Comparison
	#define LESS_SLC(a1,a2,lc) (lc = a1-a2)<0
	#define LESSEQ_SLC(a1,a2,lc) (lc = a1-a2)<=0
	#define MORE_SLC(a1,a2,lc) (lc = a1-a2)>0
	#define MOREEQ_SLC(a1,a2,lc) (lc = a1-a2)>=0
	#define EQUAL_SLC(a1,a2,lc) (lc = a1-a2)==0
	#define EQUAL_LC(lc) lc==0
	#define MORE_SLC_NOOP(a1,a2,lc) lc = a1-a2

	#define _CALL_CBFUNC
	#define _CALL_CBFUNC_COMP
	#define _DEFINE_CBFUNC
#endif

typedef struct
{
	int *pbottom;
	int top;
} sstack;

class Sort
{	
public:
	__declspec(thread) static sstack Sort::stack1;

	static void Sort::End();

	inline static void Sort::Sort_2(Elem ar[] _DEFINE_CBFUNC);
	inline static void Sort::Sort_3(Elem ar[] _DEFINE_CBFUNC);
	inline static void Sort::Sort_4(Elem ar[] _DEFINE_CBFUNC);
	inline static short Sort::MedianOf3S(Elem ar[], int a0, int a1, int a2, Elem &piv _DEFINE_CBFUNC);
	inline static short Sort::MedianOf5S(Elem ar[], int a0, int a1, int a2, int a3, int a4, Elem &piv _DEFINE_CBFUNC);
	inline static short Sort::MedianOf3NS(Elem ar[], int a0, int a1, int a2, int &ipiv _DEFINE_CBFUNC);
	inline static int Sort::MedOfThreeNoSort(Elem ar[], int a, int b, int c _DEFINE_CBFUNC);
	inline static short Sort::MedianOf5NS(Elem ar[], int a0, int a1, int a2, int a3, int a4, int &ipiv _DEFINE_CBFUNC);

	static void Sort::InitStack(sstack &stackv);
	static void Sort::InitTripleStateQuickSortTar(size_t size);
	static void Sort::FreeTripleStateQuickSortStorage();
	static void Sort::TripleStateQuickSort(Elem ar[],int a, int b _DEFINE_CBFUNC);
	inline static void Sort::_TripleStateQuickSort(Elem ar[],int a, int b _DEFINE_CBFUNC);

};
