/**********************************
Triple State Quicksort (Array Based)
Design and code: Ammar Muqaddas
Version 1.0.4 - 30/Apr/2015
ammar.mu@ku.edu.kw
ammar.mu@gmail.com
http://www.solostuff.net/tsqsort/
***********************************/

#define _USE_32BIT_TIME_T

#include <time.h>
#include <iostream>
#include <algorithm>
#include "Sort.h"

using namespace std;

// For a fair test between diffrent sorting functions. Make sure you either use __forceinline or __declspec(noinline).
__declspec(noinline) int compare(const void *elem1,const void *elem2)
{	// Note: Triple State quicksort expects ZERO when the two elements are equal.
	#ifdef ELEM_IS_STRING
		return strcmp((* (Elem *) elem1).str1,(* (Elem *) elem2).str1);
	#else	
		return (* (Elem *) elem1)-(* (Elem *) elem2);
	#endif
}

__declspec(noinline) bool compare_std(Elem elem1,Elem elem2)
{	// Suitable for std::sort only. Not qsort().

	#ifdef ELEM_IS_STRING
		return (strcmp(elem1.str1,elem2.str1))<0;
	#else	
		return elem1<elem2;
	#endif
}

int main(int argc, char* argv[])
{	// This is a simplistic test meant for demonstration. More sophisticated tests has been done.
	
	clock_t tim1,tim2=0;
	unsigned int seed = 2080;
	int const size=10000,repeat=4000,arange=6000;
	int n,i,j;
	Elem *ar;
	#ifndef ses
		size_t const ses = sizeof(*ar);
	#endif

	// Initializes recursion stack.
	Sort::InitStack(Sort::stack1);

	// Allocate array
	ar = (Elem *) malloc(ses*size);

	cout << "\nSorting " << size << " random elements, " << repeat << " times.....\n";

	// std:sort()
	srand(seed);
	for (j=0; j<repeat; j++)
	{	// Generating random array
		for (i=0; i<size; i++)
			ar[i]=((double)rand()/(RAND_MAX+1))*arange;
		tim1 = clock();
		std::sort(ar,ar+size,compare_std);
		tim2 += (clock()-tim1);
	}
	cout << "std::sort() run time: " << tim2 << " ms\n";

	// Generating SAME random arrays again by seeding with the same seed.
	srand(seed); tim2=0;
	for (j=0; j<repeat; j++)
	{	// Generating random array
		for (i=0; i<size; i++)
			ar[i]=((double)rand()/(RAND_MAX+1))*arange;
		tim1 = clock();
		Sort::TripleStateQuickSort(ar,0,size-1 _CALL_CBFUNC_COMP);
		tim2 += (clock()-tim1);
	}
	cout << "Triple State Quicksort run time: " << tim2 << " ms\n";

	// Free reserved space
	Sort::End();
	free(ar); 
}

